/*===========================================================================*/
/*
 * This file is part of
 * libalinga - a C++ library for Alligned Linguistic Annotation
 *
 * Copyright (C) 2007, 2008 Elaine Tsiang YueLien
 *
 * libalinga is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, see http://www.gnu.org/licenses, or write to the
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301, USA
 *
 *===========================================================================*/
/** @file ALingA.H                                                           *
 *                                                                           *
 *  ALingA::ALingA interface                                                 *
 *                                                                           */
/*===========================================================================*/
#ifndef ALINGAINTERFACE
#define ALINGAINTERFACE


#ifdef __GNUG__
#pragma interface
#endif

namespace ALingA
{
  class ALingA;
}

#include <ALingA/Manifold.H>
#include <ALingA/Metas.H>
#include <ALingA/Signal.H>
#include <ALingA/LingA.H>

#include <Ogg/Ogg.H>

namespace ALingA
{
  class ALingAMuxer;

  class ALingA : public Ogg::Logical
  {
    friend class LingA;

  private:
    void *			impl_;

    ALingA(const ALingA &);

    ALingA &
    operator=(const ALingA &);

    void
    add(
	LingA &
	);

    void
    remove(
	   LingA &
	   );

  public:
    /// @brief Construct an ALingA multistream encoder.
    ALingA(
	   ALingAMuxer &
	   ,long 	track
	   );

    /// @brief Construct an ALingA multistream decoder.
    ALingA(
	   Ogg::Transport &
	   );

    virtual
    ~ALingA();

    /// @brief Add an encoder with the same Transport
    ///
    /// Returns the Signal for user loading of codecType, href and Metas.
    /// Call this before ALingA writes its headers.
    Signal &
    add(
	const Ogg::Logical &		encoder
	,const Manifold &
	);

    /// @brief Add a Signal that will not be interleaved.
    ///
    /// The track number must be negative.
    /// Returns the Signal for user loading of codecType, href and Metas.
    /// The user should not leave href null.
    /// Call this before ALingA writes its headers.
    Signal &
    add(
	const long		track	
	,const Manifold &
	);

    void
    remove(
	   const Ogg::Logical &		encoder
	   );

    void
    remove(
	   const long		track	
	   );
		
    /// @brief track of ALingA stream being read or written
    ///
    /// Is same as Ogg::Logical::serialNo()
    long
    track() const
    {
      return(serialNo());
    }


    /// @brief Return the base manifold of this ALingA
    ///
    /// ALingA forms its manifold from at least one signal's manifold.
    /// The base manifold's perTime and perSpace are multiples of
    /// the perTime and perSpace of each of the signals.
    const Manifold &
    manifold() const;

    /// @brief Writes out the ALingA headers within the ALingA thread.
    ///
    /// After all encoders have called Signal::beginWrite.
    /// It will wait around until all encoders have called Signal::endWrite.
    /// While the ALingA thread persists, other threads can access its
    /// Manifold and Signal instances for purposes of coordinate transformations.
    void
    write();

    bool
    selectCallback(
		   Ogg::Packet &
		   );

    /// @brief Reads the ALingA headers within the ALingA thread.
    ///
    /// It will wait around until all encoders have called Signal::endRead.
    /// While the ALingA thread persists, other threads can access its
    /// Manifold and Signal instances for purposes of coordinate transformations.
    void
    read();

    /// @brief Return the next Signal read by ALingA thread.
    ///
    /// Call this (from a controller thread)
    /// to get the information about the Signals in the multistream
    /// that have been read so far.
    /// If the header interleave order is correct (see the ALingA.dtd
    /// for what this order is), then the application can determine
    /// the codec of a logical stream from the Signal returned by this
    /// method and instantiate it. The codec can then select and claim
    /// the logical stream.
    /// Only one iteration through the signals.
    /// Returns null if all Signals have been returned already.
    Signal *
    signal();

    /// @brief Returns parsed linga count
    ///
    /// A controller process can call this for purposes of spawning LingA threads.
    /// Blocks until the count is parsed.
    size_t
    lingaCount();

    /// @brief Returns parsed signal count
    ///
    /// A controller process can call this for purposes of spawning decoding Logical threads.
    /// Blocks until the count is parsed.
    size_t
    signalCount();

    /// @brief Returns LingAs associated with this ALingA as of time of invocation
    size_t
    countLingAs() const;

    /// @brief Returns signals associated with this ALingA as of time of invocation
    size_t
    countSignals() const;

    /// @brief Returns Signal corresponding to track
    ///
    /// Call this to get the Signal after reading of the track
    /// has already been begun. Otherwise, this may throw NoSignalAtTrack.
    Signal &
    signal(
	   long 	track
	   ) const;

    /// @brief resets for reading or writing a new ALingA
    void
    reset();

    /*--- Exceptions ---------------------------------------------------------------*/
    class InconsistentDTDversion : public Ogg::Exception
    {
    public:
      InconsistentDTDversion(
			     unsigned short	readMajor
			     ,unsigned short	readMinor
			     ,unsigned short	dtdMajor
			     ,unsigned short	dtdMinor
			     ) throw()
	: Exception("ALingA Stream's version ")
      {
	*this << readMajor << "." << readMinor
	      << " does not match ALingA implmentation's "
	      << dtdMajor << "." << dtdMinor << std::endl;
      }


      InconsistentDTDversion(
			     const InconsistentDTDversion & ex
			     ) throw()
	: Exception(ex)
      {}

      ~InconsistentDTDversion() throw()
      {}
    };
    
    class BadNonInterleavedTrackNumber : public Ogg::Exception
    {
    public:
      BadNonInterleavedTrackNumber() throw()
      : Exception("Non interleaved track number must be negative")
      {
	*this << std::endl;
      }

      BadNonInterleavedTrackNumber(
				   const BadNonInterleavedTrackNumber & ex
				   ) throw()
	: Exception(ex)
      {}
      
      ~BadNonInterleavedTrackNumber() throw()
      {}
    };

    class NotYetManifold : public Ogg::Exception
    {
    public:
      NotYetManifold() throw()
      : Exception("ALingA does not have a manifold yet")
      {
	*this << std::endl;
      }

      NotYetManifold(
		     const NotYetManifold & ex
		     ) throw()
	: Exception(ex)
      {}
      
      ~NotYetManifold() throw()
      {}
    };

    class NoSignalAtTrack : public Ogg::Exception
    {
    public:
      NoSignalAtTrack(
		      const long	track
		      ) throw()
	: Exception("No Signal at track ")
      {
	*this << track << std::endl;
      }


      NoSignalAtTrack(
		      const NoSignalAtTrack & ex
		      ) throw()
	: Exception(ex)
      {}
      
      ~NoSignalAtTrack() throw()
      {}
    };
    
  };

  /// derivable
  class ALingAMuxer : public Ogg::Transport
  {
    friend class	ALingA;

  private:
    ALingA *		alinga_;

  public:
    ALingAMuxer()
      : Transport(true)
    {}

    void
    alinga(ALingA & ala)
    {
      alinga_ = &ala;
    }

    bool
    laterThan(
	      const Ogg::Page &		p1
	      ,const Ogg::Page &	p2
	      )
    {
      if ( p1.isHeader() != p2.isHeader() )
	return( !p1.isHeader() );
      
      if ( p1.isHeader() )
	{
	  if ( (p1.pageNo() <= 1)
	       &&
	       (p2.pageNo() <= 1)
	       )
	    {
	      // both page <= 1
	      if ( p1.pageNo() != p2.pageNo() )
		return ( ! ( p1.pageNo() < p2.pageNo() ) );
	      return( ! ( p1.serialNo() < p2.serialNo() ) );
	    }
	  else
	    {
	      // at least one page > 1
	      // This is when all the page 0 headers have been written
	      // and page 1 of the smallest serialNo, which is
	      // the first ALingA Signal header page has been written.
	      // We want to write out all the ALingA Signal headers first,
	      // except for the ALingA ending header page.
	      if ( ( ( p1.serialNo() < p2.serialNo() )
		     &&
		     ( size_t(p1.pageNo()) <= alinga_->countSignals() )
		     )
		   ||
		   ( ( p2.serialNo() < p1.serialNo() )
		     &&
		     ( size_t(p2.pageNo()) <= alinga_->countSignals() )
		     )
		   )
		{
		  // write out all the Signal header pages
		  return( ! ( p1.serialNo() < p2.serialNo() ) );
		}
	      // Done all the Signal header pages
	      // We want the last ALingA header page to be the last header page
	      return( p1.serialNo() < p2.serialNo() );
	    }
	}
      // Done all the headers, go by granulePosition
      return( !( p1.granulePosition() < p2.granulePosition() ) );
    }

  };

}

#endif
