/*===========================================================================*/
/*
 * This file is part of
 * libalinga - a C++ library for Alligned Linguistic Annotation
 *
 * Copyright (C) 2007, 2008 Elaine Tsiang YueLien
 *
 * libalinga is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, see http://www.gnu.org/licenses, or write to the
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301, USA
 *
 *===========================================================================*/
/** @file LingA.H                                                            *
 *                                                                           *
 *  ALingA::LingA interface                                                  *
 *                                                                           */
/*===========================================================================*/
#ifndef	LingAInterface
#define	LingAInterface

#ifdef __GNUG__
#pragma interface
#endif

namespace ALingA
{
  class	LingA;
}

#include	<ALingA/ALingA.H>
#include	<ALingA/NodesArcs.H>
#include	<Ogg/Logical.H>


namespace	ALingA
{
  Arcs &
  join(Arcs &, Arcs &);

  class LingA : public Ogg::Logical
  {
    friend class	Node;
    friend class	Arc;

  private:
    void *			impl_;
  
    LingA(const LingA &);

    LingA &
    operator=(const LingA &);

  public:
    /// @brief Add a LingA encoder to an ALingA
    ///
    /// Specify the signal track this LingA is tied to
    LingA(
	  ALingA &
	  ,long		track
	  ,long		signalTrack
	  );

    /// @brief Add a LingA decoder to an ALingA
    LingA(
	  ALingA &
	  );

    virtual
    ~LingA();

    long
    track() const
    {
      return(serialNo());
    }


    ALingA &
    alinga() const;
    
    /// @brief Metas of the LingA stream being written or read
    ///
    /// For writing, the Metas can be modified before writing.
    Metas &
    metas();

    /// @brief Signal the LingA stream is based on
    ///
    Signal const &
    signal();

    /// @brief Write existing nodes up to given time.
    ///
    /// Each LingA can have only one stream at any given time.
    void
    write(
	  Position	time
	  );

    /// @brief End a LingA stream.
    void
    end();

    bool
    selectCallback(
		   Ogg::Packet &
		   );

    /// @brief Read the next node
    /// 
    /// Requires the instantiation of a first dummy arc.
    /// The first arc may have nodes that are unanchored.
    /// The arc and its nodes will be overwritten.
    /// The dummy arc (and nodes) are needed so new arcs
    /// and nodes may be cloned from them.
    void
    node(Arc &);

    /// @brief erase all nodes/arcs
    ///
    /// Then one can read in a fresh LingA stream.
    void
    erase();

    /// @brief augment this LingA with another
    LingA &
    operator +=(LingA &); 

    Arcs
    arcsCrossing(Epoch &);

    Nodes
    nodesIn(Epoch &);

    Arcs
    sequenceIn(Epoch &
	       , Arc *
	       );

    Arcs
    sequenceContainedIn(Epoch &
			);

    Arcs
    includeArcInSequence(Arc &
			 , Arcs &
			 , Arc * hint
			 );

    Arcs
    arcsAt(Epoch &);

    class NoNodeOrArcToClone : public Ogg::Exception
    {
    public:
      NoNodeOrArcToClone() throw()
      : Exception("No Node or Arc exists to clone from")
      {
	*this << std::endl;
      }


      NoNodeOrArcToClone(
			 const NoNodeOrArcToClone & ex
			 ) throw()
	: Exception(ex)
      {}

      ~NoNodeOrArcToClone() throw()
      {}
    };
    
    class NodeHasNoId : public Ogg::Exception
    {
    public:
      NodeHasNoId() throw()
      : Exception("Parsed Node has no id")
      {
	*this << std::endl;
      }


      NodeHasNoId(
		  const NodeHasNoId & ex
		  ) throw()
	: Exception(ex)
      {}

      ~NodeHasNoId() throw()
      {}
    };
    
    class ArcHasNoId : public Ogg::Exception
    {
    public:
      ArcHasNoId() throw()
      : Exception("Parsed Arc has no id")
      {
	*this << std::endl;
      }


      ArcHasNoId(
		 const ArcHasNoId & ex
		 ) throw()
	: Exception(ex)
      {}

      ~ArcHasNoId() throw()
      {}
    };
    
  };

}
#endif
