/*===========================================================================*/
/*
 * This file is part of
 * libalinga - a C++ library for Alligned Linguistic Annotation
 *
 * Copyright (C) 2007, 2008 Elaine Tsiang YueLien
 *
 * libalinga is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, see http://www.gnu.org/licenses, or write to the
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301, USA
 *
 *===========================================================================*/
/** @file Manifold.H                                                         *
 *                                                                           *
 *  ALingA::Manifold interface                                               *
 *                                                                           */
/*===========================================================================*/
#ifndef	ManifoldInterface
#define	ManifoldInterface

#ifdef __GNUG__
#pragma interface
#endif

#include <cstddef>

namespace ALingA
{
  /// @brief Space time co-ordinate template class
  ///
  /// @ingroup derivables
  template<class T>
  class CoOrd
  {
  public:
    T time;
    T space;

  public:
    CoOrd(
	  T	tm
	  ,T	sp
	  )
      : time(tm)
      , space(sp)
    {}
  };
  
  /// @brief ALingA::Position is unsigned.
  ///
  /// As opposed to Ogg::Position, which is signed.
  typedef 	size_t			Position;
  typedef	CoOrd<size_t>		Decimation;
  typedef 	CoOrd<Position>		Coordinate;


  class Manifold;
}

#include <ALingA/Signal.H>
#include <ALingA/XMLwriter.H>

namespace ALingA
{
  /// @brief Manifold attributes entity
  ///
  /// See http:///www.ihear.com/dtds/ALingA/0.1/ManifoldAttrs.entity
  class Manifold
  {
  private:
    void *	impl_;

  public:
    void *
    impl() const
    {
      return impl_;
    }

    Manifold(
	     const size_t		perTime
	     ,const size_t		perSpace = 1
	     ,const Glib::ustring	timeUnit = "Sec"
	     ,const Glib::ustring	spaceUnit = ""
	     ,const Position		timeBias = 0
	     ,const Position		spaceBias = 0
	     ,const size_t		gransPerFrame = 1
	     ,const Glib::ustring	timeOrigin = ""
	     );

    Manifold(
	     const Manifold &
	     ,const Signal &
	     );

    Manifold(
	     const Manifold &
	     );

    /// @brief Construct manifold from perTime attribute
    ///
    /// Throws a fit if attribute is not perTime
    Manifold(
	     const Attribute&
	     );

    virtual
    ~Manifold();

    /// @brief Set Manifold property according to attribute
    ///
    /// Throws exception if attribute not valid
    void
    operator ()(
	       const xmlpp::SaxParser::Attribute & attribute
	       );
    
    Manifold &
    operator =(const Manifold &);

    /// @brief If this manifold can be a base of the argument.
    ///
    /// Assuming identity of units and identity of time origin,
    /// the base bias must lead to inclusion of more samples
    /// and the base sampling must be of a higher resolution by integral multiples.
    /// Returns the integral multiples as Decimation.
    /// Throws exceptions otherwise.
    Decimation
    operator / (const Manifold&) const;

    /// @brief Swap manifolds if the argument is a manifold of this manifold.
    ///
    /// Do not swap if this manifold is a base of the argument.
    /// Throws exceptions if neither can be a base of the other.
    bool
    swap(
	 Manifold & manifold
	 );

    /// @brief Returns manifold's attributes as list.
    Attributes
    attributes() const;

    size_t
    perTime() const;

    size_t
    perSpace() const;

    Glib::ustring
    timeUnit() const;

    Glib::ustring
    spaceUnit() const;
    
    Position
    timeBias() const;

    Position
    spaceBias() const;

    size_t
    gransPerFrame() const;

    Glib::ustring
    timeOrigin() const;

    //--Exceptions---------------------------------------------------------------------------

    /// @brief Exception thrown by ALingA::Manifold(const Attribute&)
    class LoneAttributeNotPerTime : public Ogg::Exception
    {
    public:
      LoneAttributeNotPerTime(
			      const Glib::ustring
			      ) throw()
	;

      LoneAttributeNotPerTime(
			      const LoneAttributeNotPerTime & ex
			      ) throw()
	: Exception(ex)
      {}

      ~LoneAttributeNotPerTime() throw()
      {}
    };

    /// @brief Exception thrown by ALingA::Manifold::operator()(...)
    class InvalidAttribute : public Ogg::Exception
    {
    public:
      InvalidAttribute(
		       const Glib::ustring
		       ) throw()
	;

      InvalidAttribute(
		       const InvalidAttribute & ex
		       ) throw()
	: Exception(ex)
      {}

      ~InvalidAttribute() throw()
      {}
    };

    /// @brief Exception thrown by ALingA::Manifold::operator / (const Manifold&) const
    class OrderIndeterminate : public Ogg::Exception
    {
    public:
      OrderIndeterminate(
			 const Glib::ustring
			 ,const Glib::ustring
			 ) throw()
	;

      OrderIndeterminate(
			 const OrderIndeterminate & ex
			 ) throw()
	: Exception(ex)
      {}

      ~OrderIndeterminate() throw()
      {}
    };

    /// @brief Exception thrown by ALingA::Manifold::operator / (const Manifold&) const
    class NotBase : public Ogg::Exception
    {
    public:
      NotBase(
	      const ::ALingA::Attribute
	      ,const ::ALingA::Attribute
	      ) throw()
	;

      NotBase(
	      const NotBase & ex
	      ) throw()
	: Exception(ex)
      {}

      ~NotBase() throw()
      {}
    };

    /// @brief Exception thrown by ALingA::Manifold::operator / (const Manifold&) const
    class IncompatibleManifolds : public Ogg::Exception
    {
    public:
      IncompatibleManifolds(
			    const char *	what1
			    ,const char *	what2
			    ) throw()
	;

      IncompatibleManifolds(
			    const IncompatibleManifolds & ex
			    ) throw()
	: Exception(ex)
      {}

      ~IncompatibleManifolds() throw()
      {}
    };
  };
}

#endif

