/*===========================================================================*/
/*
 * This file is part of
 * libalinga - a C++ library for Alligned Linguistic Annotation
 *
 * Copyright (C) 2007, 2008 Elaine Tsiang YueLien
 *
 * libalinga is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, see http://www.gnu.org/licenses, or write to the
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301, USA
 *
 *===========================================================================*/
/** @file NodesArcs.H                                                        *
 *                                                                           *
 *  ALingA::Node                                                             *
 *  ALingA::Arc                                                              *
 *  ALingA::Nodes                                                            *
 *  ALingA::Arcs                                                             *
 *                                                                           */
/*===========================================================================*/
#ifndef	NodesArcsInterface
#define	NodesArcsInterface

#ifdef __GNUG__
#pragma interface
#endif

#include	<set>

namespace	ALingA
{
  typedef unsigned int		Id;

  class Epoch;
  class Node;
 
  class	Label;
  class Arc;

  bool
  operator < (const Node &, const Node &);

  bool
  operator < (const Arc &, const Arc &);

  class PrecedingNode
  {
  public:
    bool operator ()(Node * const & node1, Node * const & node2) const
    {
      return( *node1 < *node2 );
    }
  };
  typedef std::multiset< Node *, PrecedingNode >	Nodes;
  typedef Nodes::iterator				iNodes;

  class PrecedingArc
  {
  public:
    bool operator ()(Arc * const & arc1, Arc * const & arc2) const
    {
      return( *arc1 < *arc2 );
    }
  };
  typedef std::multiset< Arc *, PrecedingArc >		Arcs;
  typedef Arcs::iterator				iArcs;
  
}

#include	<ALingA/LingA.H>

namespace	ALingA
{

  /// @brief Left is beginning time, right is ending time
  ///
  /// The Manifold biases, however, may take on negative values.
  struct Epoch
  {
    /// @brief Exception thrown by Epoch constructor
    class Invalid : public Ogg::Exception
    {
    public:
      Invalid(
	      const Position	time1
	      , const Position	time2
	      ) throw()
	: Exception("Not an epoch ")
      {
	*this << time1 << " > " << time2 << std::endl;
      }

      Invalid(
	      const Invalid & ex
	      ) throw()
	: Exception(ex)
      {}
      
      ~Invalid() throw()
      {}
    };

    Position	left;
    Position	right;

    Epoch(Position l
	  ,Position r
	  )
      : left(l)
      , right(r)
    {
      if ( left > right )
	throw Invalid(l,r);
    }

  };

  class EpochRelation
  {
  public:
    enum Type
      {
	BEFORE = -3
	,LEFT = -2
	,CONTAINED = -1
	,EQUAL = 0
	,CONTAINING = 1
	,RIGHT = 2
	,AFTER = 3
      };

    Type
    operator ()(const Epoch &e1, const Epoch &e2) const
    {
      if ( e1.right < e2.left )
	return BEFORE;

      if ( e2.right < e1.left )
	return AFTER;

      // e1.right >= e2.left and e2.right >= e1.left

      if ( e1.left < e2.left )
	{
	  if ( e1.right <= e2.right )
	    return LEFT;
	  else // e1.right > e2.right
	    return CONTAINING;
	}
	else if ( e1.left == e2.left )
	  {
	    if ( e1.right < e2.right )
	      return LEFT;
	    else if ( e1.right == e2.right )
	      return EQUAL;
	    else // e1.right > e2.right
	      return RIGHT;
	  }
	else // e1.left > e2.left
	  {
	    if ( e1.right < e2.right )
	      return CONTAINED;
	    else // ( e1.right >= e2.right )
	      return RIGHT;
	  }
      
    }
  };

  inline
  bool
  operator <(const Epoch & e1, const Epoch & e2)
  {
    EpochRelation r;
    if ( int(r(e1,e2)) < 0 )
      return true;
    else
      return false;
  }

  struct Box
  {
  public:
    Position	left;
    Position	right;

    Box(Position l
	,Position r
	)
      : left(l)
      , right(r)
    {}
  };

  Arcs
  path(Arc &, Arc &);

  class Node
  {
    friend class LingA;
    friend class Arc;

    friend bool 	operator < (const Node &, const Node &);
    friend Arcs 	path(Arc &, Arc &);

  private:
    void *		impl_;

    Node(const Node &);

    Node &
    operator =(const Node &);

  public:
    void *		impl();

  public:
    /// @brief Exception thrown by Node::time()
    class TimeUndefined : public Ogg::Exception
    {
    public:
      TimeUndefined() throw()
      : Exception("Node time undefined ")
      {}

      TimeUndefined(
		    const TimeUndefined & ex
		    ) throw()
	: Exception(ex)
      {}
      
      ~TimeUndefined() throw()
      {}
    };

    /// @brief Exception thrown by Node::time()
    class SpaceUndefined : public Ogg::Exception
    {
    public:
      SpaceUndefined() throw()
      : Exception("Node space undefined ")
      {}

      SpaceUndefined(
		     const SpaceUndefined & ex
		     ) throw()
	: Exception(ex)
      {}
      
      ~SpaceUndefined() throw()
      {}
    };

    Node(
	 LingA &	linga
	 );

    Node(
	 Id		id
	 ,LingA &	linga
	 );

    Node(
	 LingA &	linga
	 ,Position	time
	 );

    Node(
	 LingA &	linga
	 ,Coordinate
	 );

    virtual
    ~Node();

    Id
    id() const;

    LingA &
    linga();

    bool
    time(Position time);

    bool
    space(Position space);

    /// @brief Return node time if defined
    ///
    /// May throw TimeUndefined
    Position
    time() const;

    /// @brief Return node space if defined
    ///
    /// May throw SpaceUndefined
    Position
    space() const;

    Arcs
    lefts() const;

    Arcs
    rights() const;

    /// @brief If this is a "junk" node that results from a defective LingA stream
    bool
    junk();

    bool
    operator ==(const Node &) const;

    /// @brief Clone a bare node
    ///
    /// Override this with derived types.
    virtual
    Node &
    clone(Id id)
    {
      Node & n = * new Node(id
			    ,linga()
			    );

      return(n);
    }

  };

  struct Label
  {
    Glib::ustring	randomVar;
    Glib::ustring	value;
    Glib::ustring	likelihood;

    Label(
	  Glib::ustring		r
	  ,Glib::ustring	v
	  ,Glib::ustring	l
	  )
      : randomVar(r)
      , value(v)
      , likelihood(l)
    {}

  };

  inline
  bool
  operator < (const Label & l1, const Label & l2)
  {
    if (  l1.randomVar != l2.randomVar )
      return( l1.randomVar < l2.randomVar );
    else
      return( l1.value < l2.value );
  }

  class Arc
  {
    friend class LingA;

    friend bool operator < (const Arc &, const Arc &);
  private:
    void *		impl_;

    Arc(const Arc &);

    Arc &
    operator =(const Arc &);

  public:
    void *		impl();

  public:
    Arc(
	Node &		left
	,Node &		right
	,Label		label
	,Id		id
	);

    Arc(
	Node &		left
	,Node &		right
	,Label		label
	);

    virtual
    ~Arc();
    
    Id
    id() const;

    /// @brief Get Metas of Arc
    ///
    /// The application stocks the Metas directly.
    Metas &
    metas();

    /// @brief	Epoch from arc's left node's time to right node's time.
    ///
    /// May throw EpochUndefined.
    Epoch
    epoch() const;

    /// @brief	Epoch from arc's left node's space to right node's space.
    ///
    /// May throw BoxUndefined.
    Box
    box() const;
    
    /// @brief Change Arc's left node
    ///
    /// Returns true for success.
    bool
    left(Node &);

    Node &
    left() const;

    /// @brief Change Arc's right node
    ///
    /// Returns true for success.
    bool
    right(Node &);

    Node &
    right() const;

    void
    label(Label &);
      
    Label
    label() const;

    /// @brief If this is a "junk" arc that results from a defective LingA stream
    bool
    junk();

    bool
    operator ==(const Arc &) const;
    
    /// @brief Clone a bare arc
    ///
    /// Override this with derived types.
    virtual
    Arc &
    clone(Id id)
    {
      Arc & n = * new Arc(
			  left()
			  ,right()
			  ,label()
			  ,id
			  );
      return(n);
    }
  };

  
}

#endif  
