/*===========================================================================*/
/*
 * This file is part of
 * libalinga - a C++ library for Alligned Linguistic Annotation
 *
 * Copyright (C) 2007, 2008 Elaine Tsiang YueLien
 *
 * libalinga is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, see http://www.gnu.org/licenses, or write to the
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301, USA
 *
 *===========================================================================*/
/** @file XMLparser.H                                                         *
 *                                                                           *
 *  ALingA::XMLparser interface                                               *
 *                                                                           */
/*===========================================================================*/
#ifndef	XMLparserInterface
#define	XMLparserInterface

#ifdef __GNUG__
#pragma interface
#endif

#include <libxml++/libxml++.h>
#include <glibmm/ustring.h>

#include <Ogg/Exception.H>

namespace	ALingA
{
  /// @brief ALingA XML parser base class
  ///
  /// @ingroup derivables
  class XMLparser
  {
  private:
    void *		impl_;

  public:
    XMLparser(
	      const Glib::ustring	elementName
	      ,XMLparser * 		= 0	// parent
	      );

    virtual
    ~XMLparser();

    const Glib::ustring &
    elementName() const;

    void
    parse(
	  void *	data
	  ,size_t	size
	  );

    XMLparser &
    top();

    // the on_ methods may be variously restricted with exception throwing
    virtual void on_start_document();
    virtual void on_end_document();
    // start_element is handling a contained element, which may vary
    // it tests the arg name for which element is this current invocation,
    // derived class: accesses or creates the pointer to the receiving data area,
    // and creates the parser with that pointer and the attributes
    virtual
    void
    on_start_element(
		     const Glib::ustring& name
		     , const xmlpp::SaxParser::AttributeList& attributes
		     );

    // looks at the stack top parser, and verifies that corresponds
    // to the contained element of the passed name,
    // otherwise throw BadEndElement
    // deletes the stack top parser (which causes it to be popped)
    // which is the parser for the contained element
    // which is, counter-intuitively, this parser.
    virtual
    void
    on_end_element(const Glib::ustring& name);

    // ALingA serialization mandates that PCDATA does not cross packet boundaries
    virtual void on_characters(const Glib::ustring& characters);
    virtual void on_cdata_block(const Glib::ustring& text);


    //--Exceptions---------------------------------------------------------------------------

    /// @brief Exception thrown by ALingA::XMLparser::parse(..)
    class UnexpectedDocument : public Ogg::Exception
    {
    public:
      UnexpectedDocument() throw()
	;

      UnexpectedDocument(
			 const UnexpectedDocument & ex
			 ) throw()
	: Exception(ex)
      {}

      ~UnexpectedDocument() throw()
      {}
    };

    /// @brief Exception thrown by ALingA::XMLparser()::parse(..)
    class UnexpectedElement : public Ogg::Exception
    {
    public:
      UnexpectedElement(
			const Glib::ustring
			) throw()
	;

      UnexpectedElement(
			const UnexpectedElement & ex
			) throw()
	: Exception(ex)
      {}

      ~UnexpectedElement() throw()
      {}
    };

    /// @brief Exception thrown by ALingA::XMLparser()::parse(..)
    class UnexpectedAttribute : public Ogg::Exception
    {
    public:
      UnexpectedAttribute(
			  const Glib::ustring
			  ) throw()
	;

      UnexpectedAttribute(
			  const UnexpectedAttribute & ex
			  ) throw()
	: Exception(ex)
      {}

      ~UnexpectedAttribute() throw()
      {}
    };

    /// @brief Exception thrown by ALingA::XMLparser()::parse(..)
    class UnexpectedCharacters : public Ogg::Exception
    {
    public:
      UnexpectedCharacters(
			   const Glib::ustring
			   ) throw()
	;

      UnexpectedCharacters(
			   const UnexpectedCharacters & ex
			   ) throw()
	: Exception(ex)
      {}

      ~UnexpectedCharacters() throw()
      {}
    };
    
    /// @brief Exception thrown by ALingA::XMLparser()::parse(..)
    class UnexpectedCData : public Ogg::Exception
    {
    public:
      UnexpectedCData(
		      const Glib::ustring
		      ) throw()
	;

      UnexpectedCData(
		      const UnexpectedCData & ex
		      ) throw()
	: Exception(ex)
      {}

      ~UnexpectedCData() throw()
      {}
    };
    
    /// @brief Exception thrown by ALingA::~XMLparser()
    class ContextCorruption : public Ogg::Exception
    {
    public:
      ContextCorruption(
			const Glib::ustring
			,const Glib::ustring
			) throw()
	;

      ContextCorruption(
			const ContextCorruption & ex
			) throw()
	: Exception(ex)
      {}

      ~ContextCorruption() throw()
      {}
    };

    /// @brief Exception thrown by ALingA::~XMLparser()
    class NoContext : public Ogg::Exception
    {
    public:
      NoContext(
		const Glib::ustring
		) throw()
	;

      NoContext(
		const NoContext & ex
		) throw()
	: Exception(ex)
      {}

      ~NoContext() throw()
      {}
    };

    /// @brief Exception thrown by ALingA::XMLparser()::parse(..)
    class BadAttributeValue : public Ogg::Exception
    {
    public:
      BadAttributeValue(
			const xmlpp::SaxParser::Attribute &
			) throw()
	;

      BadAttributeValue(
			const BadAttributeValue & ex
			) throw()
	: Exception(ex)
      {}

      ~BadAttributeValue() throw()
      {}
    };

  };

}
#endif
