/*===========================================================================*/
/*
 * This file is part of
 * libalinga - a C++ library for Alligned Linguistic Annotation
 *
 * Copyright (C) 2007, 2008 Elaine Tsiang YueLien
 *
 * libalinga is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, see http://www.gnu.org/licenses, or write to the
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301, USA
 *
 *===========================================================================*/
/** @file XMLwriter.H                                                        *
 *                                                                           *
 *  ALingA::XMLwriterinterface                                               *
 *                                                                           */
/*===========================================================================*/
#ifndef	XMLwriterInterface
#define	XMLwriterInterface

#ifdef __GNUG__
#pragma interface
#endif

#include <libxml++/libxml++.h>
#include <glibmm/ustring.h>

#include <Ogg/Exception.H>

namespace	ALingA
{
  class Attribute;
  typedef xmlpp::SaxParser::AttributeList		Attributes;
}

#include <ALingA/Manifold.H>
#include <ALingA/Metas.H>

#include <sstream>
/**********************************************************************
 * Where Glib::ustring locale_to_utf8(const std::string&) is declared *
 **********************************************************************/
#include <glibmm/convert.h>


namespace	ALingA
{
  using	std::ostringstream;

  /// @brief Attribute base class
  class Attribute : public xmlpp::SaxParser::Attribute
  {
  public:
    Attribute(
	      Glib::ustring const &	nm
	      )
      : xmlpp::SaxParser::Attribute(
				    nm
				    ,""
				    )
    {}

    virtual
    ~Attribute()
    {}

    virtual
    void *
    typedValue() const
    {
      return(0);
    }

  };

  /// @brief template class for a typed attribute
  ///
  /// ostringstream operator << (ostringstream &, T&) must be defined
  template<class T>
  class TypedAttribute : public Attribute
  {
  private:
    T	t_;
  public:
    TypedAttribute(
		   Glib::ustring const &	nm
		   , T				t
		   )
      : Attribute(nm)
      , t_(t)
    {
      ostringstream os;
      os << t;
      value = Glib::locale_to_utf8(os.str());
    }

    TypedAttribute(
		   const char * &		nm
		   , T				t
		   )
    {
      TypedAttribute(
		     Glib::ustring(nm)
		     ,t
		     );
    }

    T *
    typedValue()
    {
      return &t_;
    }
  };

  typedef TypedAttribute<unsigned short>		UShortAttribute;
  typedef TypedAttribute<long>				LongAttribute;
  typedef TypedAttribute<long long>			LongLongAttribute;
  typedef TypedAttribute<size_t>			SizeAttribute;

  /// @brief Base class for expressing objects in XML
  ///
  class XMLwriter
  {
  private:
    void *	impl_;

  public:
    XMLwriter(
	      Glib::ustring	rootName
	      ,unsigned short	dtdMajor
	      ,unsigned short	dtdMinor
	      );

    XMLwriter(
	      XMLwriter &	parent
	      ,Glib::ustring	elementName
	      ,bool		isInstruction = false
	      );

    virtual
    ~XMLwriter();

    /// @brief Express attribute in XML
    void
    attribute(
	      Glib::ustring	name
	      ,Glib::ustring	value
	      );
    /// @brief Express end of attributes
    void
    endAttributes();
		  
    /// @brief Express #PCDATA
    void
    characters(
	       Glib::ustring &
	       );

    /// @brief Express commonly encountered ALingA Manifold attributes
    void
    manifold(
	     Manifold &		manifold
	     );

    /// @brief Express commonly encountered ALingA xmlns-related attributes
    void
    xmlns(
	  Glib::ustring
	  );

    /// @brief Express Metas in XML
    void
    metas(
	  Metas &
	  );

    /// @brief Size of current XML chunk
    size_t
    textSize();

    /// @brief Write current XML chunk to data area
    ///
    /// Note no overwriting safeguard
    void
    write(
	  void *	data
	  );

    //--Exceptions---------------------------------------------------------------------------

    /// @brief Exception thrown by ALingA::XMLwriter(Glib::ustring rootName,...)
    class RootNameMoreThan8Bytes : public Ogg::Exception
    {
    public:
      RootNameMoreThan8Bytes(
			     const Glib::ustring
			     ) throw()
	;

      RootNameMoreThan8Bytes(
			     const RootNameMoreThan8Bytes & ex
			     ) throw()
	: Exception(ex)
      {}

      ~RootNameMoreThan8Bytes() throw()
      {}
    }
      ;

    /// @brief Exception thrown by ALingA::~XMLwriter()
    class TagNotCompleted : public Ogg::Exception
    {
    public:
      TagNotCompleted(
		      const Glib::ustring
		      ) throw()
	;

      TagNotCompleted(
		      const TagNotCompleted & ex
		      ) throw()
	: Exception(ex)
      {}

      ~TagNotCompleted() throw()
      {}
    };

    /// @brief Exception thrown by ALingA::~XMLwriter()
    class ContextCorruption : public Ogg::Exception
    {
    public:
      ContextCorruption(
			const Glib::ustring
			) throw()
	;

      ContextCorruption(
			const ContextCorruption & ex
			) throw()
	: Exception(ex)
      {}

      ~ContextCorruption() throw()
      {}
    };

    /// @brief Exception thrown by ALingA::~XMLwriter()
    class NoContext : public Ogg::Exception
    {
    public:
      NoContext(
		const Glib::ustring
		) throw()
	;

      NoContext(
		const NoContext & ex
		) throw()
	: Exception(ex)
      {}

      ~NoContext() throw()
      {}
    };

  };

}
#endif
