/*===========================================================================*/
/*
 * This file is part of
 * libalinga - a C++ library for Alligned Linguistic Annotation
 *
 * Copyright (C) 2007, 2008 Elaine Tsiang YueLien
 *
 * libalinga is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, see http://www.gnu.org/licenses, or write to the
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301, USA
 *
 *===========================================================================*/
/** @file ALingAImpl.C                                                       *
 *                                                                           *
 *  ALingA::ALingA implementation                                            *
 *                                                                           */
/*===========================================================================*/
#ifdef __GNUG__
#pragma implementation "ALingA.H"
#pragma implementation "ALingAImpl.H"
#endif

#include	<ALingAImpl.H>
#include	<ALingA/Metas.H>
#include	<Ogg/Debug.H>

extern "C" 
{
# include	<stdlib.h>
}

namespace ALingA
{
  using Ogg::Debug;
  using Ogg::debug;

  /*------ Writing ALingA ----------------------------------------------------------*/

  ALingA::ALingA(
		 ALingAMuxer &	muxer
		 ,long		track
		 )
    : Logical(muxer
	      ,track
	      )
  {
    ALingAImpl *impl = new ALingAImpl(true);

    impl->intf_ = this;
    impl_ = impl;
  }
  
  ALingAImpl::ALingAImpl(bool writing)
    : writing_(writing)
    , xmlns_("http://www.ihear.com/dtds/ALingA.dtd")
    , manifold_(1)
    , signalCount_(0)
    , lingaCount_(0)
      // if writing, then doneParsing always true
      // else, it starts out false
    , doneParsing_(writing)
    , reader_(0)
    , signals_()
    , argSigs_(0)
    , lingas_()
  {
    pthread_mutex_init(&lockCount_, 0);
    pthread_cond_init(&condCount_, 0);

    pthread_cond_init(&condSignals_, 0);
    pthread_mutex_init(&lockSignals_, 0);
    pthread_mutex_init(&lockLingas_, 0);
  }

  ALingA::~ALingA()
  {
    if ( impl_ )
      {
	ALingAImpl * impl = static_cast<ALingAImpl *>(impl_);
	delete impl;
	impl_ = 0;
      }
  }
  
  void
  ALingA::add(
	      LingA &	linga
	      )
  {
    ALingAImpl * impl = static_cast<ALingAImpl *>(impl_);

    impl->add(linga);
  }

  void
  ALingAImpl::add(
		  LingA & linga
		  )
  {
    LingA *	la = 0;

    lockLingas();
    if ( !writing_
	 ||
	 ( lingaCount_ < 1 )
	 )
      {
	la = &linga;
	lingas_[linga.track()] = la;
      }
    unlockLingas();

    if ( !la )
      //throw ALingA::TooLateToAddLingA()
      ;
  }
 
  void
  ALingA::remove(
		 LingA &	linga
		 )
  {
    ALingAImpl * impl = static_cast<ALingAImpl *>(impl_);

    impl->remove(linga);
  }
   
  void
  ALingAImpl::remove(
		     LingA & linga
		     )
  {
    lockLingas();
    if ( writing_
	 &&
	 ( lingaCount_ < 1 )
	 )
      {
	ALingAImpl::iLings ilings = lingas_.find(linga.track());
	if ( ilings != lingas_.end() )
	  {
	    lingas_.erase(ilings);
	  }
      }
    unlockLingas();
  }

  Signal &
  ALingA::add(
	      const Ogg::Logical &	encoder
	      ,const Manifold &		manifold
	      )
  {
    ALingAImpl * impl = static_cast<ALingAImpl *>(impl_);
    
    return(impl->add(encoder.serialNo()
		     ,manifold
		     )
	   );
  }

  Signal &
  ALingA::add(
	      const long		track
	      ,const Manifold &		manifold
	      )
  {
    if ( 0 <= track )
      throw BadNonInterleavedTrackNumber();

    ALingAImpl * impl = static_cast<ALingAImpl *>(impl_);
    
    return(impl->add(track
		     ,manifold
		     )
	   );
  }

  Signal &
  ALingAImpl::add(
		  const long		track
		  ,const Manifold &	manifold
		  )
  {
    if ( !writing_ )
      //throw ALingA::AddingTrackToNonWritingAlingA()
	;

    SignalImpl * sig = 0;

    lockSignals();
    if ( signalCount_ < 1 )
      // writing has not started
      {
	iSigs isigs = signals_.find(track);
	if ( isigs == signals_.end() )
	  {
	    sig = new SignalImpl(track
				 , manifold_.gransPerFrame()
				 , Coordinate(manifold.timeBias()
					      ,manifold.spaceBias()
					      )
				 , Decimation(1,1)
				 );
	    signals_[track] = sig;
	  }
	else
	  {
	    sig = isigs->second;
	    sig->bias_ = Coordinate(manifold.timeBias()
				    ,manifold.spaceBias()
				    );
	    sig->decimation_ = Decimation(1,1);
	  }

	// save my current manifold
	Manifold	m0(manifold_);
	Manifold	m(manifold);
	if ( !manifold_.swap(m) )
	  {
	    sig->decimation_ = manifold_ / manifold;
	  }
	else
	  {
	    // recompute this signal's decimation
	    // the swap could be space only
	    sig->decimation_ = manifold_ / manifold;

	    // recompute the other signals' decimations
	    Manifold	m(1);
	    for ( isigs = signals_.begin()
		    ;
		  isigs != signals_.end()
		    ;
		  ++isigs
		  )
	      {
		if ( isigs->second != sig )
		  {
		    // m0 is my old manifold, against which the decimations were computed
		    // first recompute what the signal manifold is
		    m = Manifold(m0
				 ,*isigs->second
				 );
		    isigs->second->decimation_ = manifold_ / m;
		  }
	      }
	  }
      }
    unlockSignals();

    if ( !sig )
      //throw ALingA::TooLateToAddSignal()
      ;

    return(*sig);
  }

  void
  ALingA::remove(
		 const Ogg::Logical &	encoder
		 )
  {
    ALingAImpl * impl = static_cast<ALingAImpl *>(impl_);
    
    return(impl->remove(encoder.serialNo()));
  }

  void
  ALingA::remove(
		 const long		track
		 )
  {
    if ( 0 <= track )
      throw BadNonInterleavedTrackNumber();

    ALingAImpl * impl = static_cast<ALingAImpl *>(impl_);
    
    return(impl->remove(track));
  }

  void
  ALingAImpl::remove(
		     const long		track
		     )
  {
    if ( !writing_ )
      //throw ALingA::RemovingTrackFromNonWritingAlingA()
	;

    SignalImpl *	sig = 0;

    lockSignals();
    if ( signalCount_ < 1 )
      // ALingA is not operating writing
      {
	ALingAImpl::iSigs isigs = signals_.find(track);
	if ( isigs != signals_.end() )
	  {
	    delete (sig = isigs->second);
	    signals_.erase(isigs);
	  }
      }
    unlockSignals();

    if ( !sig )
      //throw ALingA::TooEarlyToRemoveSignal()
      ;

  }

  const Manifold &
  ALingA::manifold() const
  {
    ALingAImpl * impl = static_cast<ALingAImpl *>(impl_);
    if ( !impl->writing_ )
      impl->waitForCounts();
    return(impl->manifold_);
  }

  void
  ALingA::write()
  {
    ALingAMuxer * mux = static_cast<ALingAMuxer *>(&transport());
    
    mux->alinga(*this);
    static_cast<ALingAImpl *>(impl_)->write();
  }

  void
  ALingAImpl::write()
  {
    lockSignals();
    {
      if ( (signalCount_ = signals_.size()) < 1 )
	{
	  unlockSignals();
	  //throw ALingA::ALingAHasYetNoSignal();
	}
    }
    unlockSignals();

    lockLingas();
    {
      if ( (lingaCount_ = lingas_.size()) < 1 )
	{
	  unlockLingas();
	  //throw ALingA::ALingAHasYetNoLingA();
	}
    }
    unlockLingas();
    gotCounts();

    // once the counts are nonzero
    // its OK to access the signals without locking

    waitForSignalsReady();

    // write ALingA headers
    writeHeaders();

    waitForSignalsDone();

    // Done with this ALingA bitstream
    // Don't reset, leave to user.
    // reset();
  }


  void
  ALingAImpl::writeHeaders()
  {
    Ogg::Logical::Writer & w = intf_->writer();


    char * const	data = reinterpret_cast<char *>(w->data());
    size_t		off = 12;

    // Ogg-format preamble in header 0
    data[0] = 'A';data[1]='L';data[2]='i';data[3]='n';data[4]='g';data[5] = 'A';
    data[6]=0;data[7]=0;

    // Encode will impose big-endian order
    Encode		enc(Unsigned16Bit);
    unsigned short *	u = reinterpret_cast<unsigned short *>(data+8);
    *u = dtdMajor_;
    enc.encode(u
	       ,sizeof(unsigned short)
	       );
    ++u;
    *u = dtdMinor_;
    enc.encode(u
	       ,sizeof(unsigned short)
	       );


    XMLwriter xmlWriter(
			"ALingA"
			,dtdMajor_
			,dtdMinor_
			);

    {
      XMLwriter alingaWriter(
			     xmlWriter
			     ,"ALingA"
			     ,false	// is not instruction, libxml doesn't parse instructions
			     );
      LongAttribute	trk("track"
			    ,intf_->serialNo()
			    );
      alingaWriter.attribute(
			     trk.name
			     ,trk.value
			     );
      alingaWriter.manifold(manifold_);

      alingaWriter.xmlns(xmlns_);
      
      {
	SizeAttribute	count("signalCount"
			      ,signalCount_
			      );
	alingaWriter.attribute(
			       count.name
			       ,count.value
			       );
      }
      {
	SizeAttribute	count("lingaCount"
			      ,lingaCount_
			      );
	alingaWriter.attribute(
			       count.name
			       ,count.value
			       );
      }

      alingaWriter.endAttributes();

      w.dataSize(off+xmlWriter.textSize());
      alingaWriter.write(
			 reinterpret_cast<char *>(w->data())+off
			 );

      for ( iSigs isigs = signals_.begin()
	      ;
	    isigs != signals_.end()
	      ;
	    ++isigs
	    )
	{
	  ++w;
	  w.flush();

	  SignalImpl & sig = * isigs->second;
	
	  {
	    XMLwriter signalWriter(
				   alingaWriter
				   ,"Signal"
				   );
	    {
	      LongAttribute	trk("track"
				    ,sig.track_
				    );
	      signalWriter.attribute(
				     trk.name
				     ,trk.value
				     );
	    }
	    signalWriter.attribute(
				   "codecType"
				   ,sig.codecType_
				   );
	    if ( !sig.href_.empty() )
	      {
		signalWriter.attribute(
				       "xlink:type"
				       ,"simple"
				       );
		signalWriter.attribute(
				       "xlink:href"
				       ,sig.href_
				       );
	    
	      }
	    {
	      SizeAttribute	siz("baseGransPerFrame"
				    ,sig.baseGransPerFrame_
				    );
	      signalWriter.attribute(
				     siz.name
				     ,siz.value
				     );
	    }
	    {
	      SizeAttribute	siz("timeBias"
				    ,sig.bias_.time
				    );
	      signalWriter.attribute(
				     siz.name
				     ,siz.value
				     );
	    }
	    {
	      SizeAttribute	siz("spaceBias"
				    ,sig.bias_.space
				    );
	      signalWriter.attribute(
				     siz.name
				     ,siz.value
				     );
	    }
	    {
	      SizeAttribute	siz("timeDecimation"
				    ,sig.decimation_.time
				    );
	      signalWriter.attribute(
				     siz.name
				     ,siz.value
				     );
	    }
	    {
	      SizeAttribute	siz("spaceDecimation"
				    ,sig.decimation_.space
				    );
	      signalWriter.attribute(
				     siz.name
				     ,siz.value
				     );
	    }
	    signalWriter.endAttributes();

	    if ( !sig.metas_.empty() )
	      {
		XMLwriter metasWriter(
				      signalWriter
				      ,"Metas"
				      );
		metasWriter.endAttributes();
		metasWriter.metas(sig.metas_);
	      }
	    // Metas tag closed
	  }
	  // Signal tag closed
	  w.dataSize(alingaWriter.textSize());
	  alingaWriter.write(
			     reinterpret_cast<char *>(w->data())
			     );
	}

      ++w;
      // Ending ALingA page
      {
	XMLwriter signalWriter(
			       alingaWriter
			       ,"Signal"
			       );
	{
	  LongAttribute	trk("track"
			    ,0
			    );
	  signalWriter.attribute(
				 trk.name
				 ,trk.value
				 );
	}
	signalWriter.attribute(
			       "codecType"
			       ,"ALingA"
			       );
	signalWriter.endAttributes();

      }
      // Ending ALingA Signal tag closed
      w.dataSize(alingaWriter.textSize());
      alingaWriter.write(
			 reinterpret_cast<char *>(w->data())
			 );
    }
    // ALingA has no end tag
    // last Signal header to be sent out upon this deletion

    debug(false,
	  "ALingA aout to send the ending last header"
	  );
    delete &w;

    // Done writing ALingA headers
  }

  /*------ wait methods ------------------------------------------------------------*/

  void
  ALingAImpl::waitForCounts()
  {
    pthread_mutex_lock(&lockCount_);
    {
      while ( (signalCount_ <= 0)
	      ||
	      (lingaCount_ <=0)
	      )
	pthread_cond_wait(&condCount_, &lockCount_);
    }
    pthread_mutex_unlock(&lockCount_);
  }

  void
  ALingAImpl::gotCounts()
  {
    pthread_mutex_lock(&lockCount_);
    {
      pthread_cond_broadcast(&condCount_);
    }
    pthread_mutex_unlock(&lockCount_);
  }

  void
  ALingAImpl::waitForSignal()
  {
    // make sure we have counts
    waitForCounts();

    pthread_mutex_lock(&lockSignals_);
    {
      while( ( argSigs_ >= signals_.size() )
	     &&
	     ( signals_.size() < signalCount_ )
	     )
	pthread_cond_wait(&condSignals_, &lockSignals_);
    }
    pthread_mutex_unlock(&lockSignals_);
  }

  void
  ALingAImpl::gotSignal(SignalImpl * signalImpl)
  {
    pthread_mutex_lock(&lockSignals_);
    {
      if ( signalImpl )
	signals_[signalImpl->track_] = signalImpl;
      // signal got signal or end of signals
      pthread_cond_broadcast(&condSignals_);
    }
    pthread_mutex_unlock(&lockSignals_);
  }

  void
  ALingAImpl::waitForSignalsReady()
  {
    for ( iSigs isigs = signals_.begin()
	    ;
	  isigs != signals_.end()
	    ;
	  ++isigs
	  )
      {
	if ( 0 < isigs->first )
	  // block till non-empty codecType
	  isigs->second->codecTypeFilled();
      }
  }

  void
  ALingAImpl::waitForSignalsDone()
  {
    for ( iSigs isigs = signals_.begin()
	    ;
	  isigs != signals_.end()
	    ;
	  ++isigs
	  )
      {
	if ( 0 < isigs->first )
	  // block till empty codecType
	  isigs->second->codecTypeEmpty();
      }

    debug(true,
	  "ALingA detected all signals done"
	  );
  }
  /*------ reset reading or writing ------------------------------------------------*/
  void
  ALingA::reset()
  {
    ALingAImpl * impl = static_cast<ALingAImpl *>(impl_);

    impl->reset();
  }

  void
  ALingAImpl::reset()
  {
    doneParsing_ = false;

    lockSignals();
    signalCount_ = 0;
    for ( iSigs i = signals_.begin()
	    ;
	  i != signals_.end()
	    ;
	  ++i
	  )
      {
	// this is why we don't need to call remove on reading
	// similar to Ogg::Transport
	delete i->second;
      }
    signals_.clear();
    argSigs_ = 0;
    unlockSignals();
    
    lockLingas();
    lingaCount_ = 0;
    lingas_.clear();
    unlockLingas();

  }

  /*------ Reading ALingA ----------------------------------------------------------*/
  
  ALingA::ALingA(
		 Ogg::Transport & demuxer
		 )
    : Ogg::Logical(demuxer)
  {
    ALingAImpl *impl = new ALingAImpl(
				      false
				      );

    impl->intf_ = this;
    impl_ = impl;
  }
  
  bool
  ALingA::selectCallback(
			 Ogg::Packet &	pkt
			 )
  {
    char * header = static_cast<char *>(pkt.data());

    if( 
       !strncmp(header, "ALingA", 6)
       &&
       ( header[6] == '\0' )
       &&
       ( header[7] == '\0' )
	)
      {
	Encode			enc(Unsigned16Bit);
	unsigned short *	uMajor = reinterpret_cast<unsigned short *>(header+8);
	unsigned short *	uMinor = reinterpret_cast<unsigned short *>(header+10);
	enc.decode(uMajor
		   ,sizeof(unsigned short)
		   );
	enc.decode(uMinor
		   ,sizeof(unsigned short)
		   );
	if ( (*uMajor != ALingAImpl::dtdMajor_)
	     ||
	     (*uMinor != ALingAImpl::dtdMinor_)
	     )
	  throw ALingA::InconsistentDTDversion(
					       *uMajor
					       ,*uMinor
					       ,ALingAImpl::dtdMajor_
					       ,ALingAImpl::dtdMinor_
					       );

	return(true);
      }
    return(false);
  }

  void
  ALingA::read()
  {
    static_cast<ALingAImpl *>(impl_)->read();
  }

  Signal *
  ALingA::signal()
  {
    ALingAImpl * impl = static_cast<ALingAImpl *>(impl_);
    return(impl->signal());
  }

  Signal *
  ALingAImpl::signal()
  {
    if ( !writing_ )
      {
	waitForSignal();
	// either we have a signal
	// or we got all the promised signals

	lockSignals();
	if ( argSigs_ < signals_.size() )
	  {
	    iSigs i = signals_.begin();
	    for ( long n=0;n<argSigs_;++n)
	      ++i;
	    Signal * sig = i->second;
	    ++argSigs_;
	    unlockSignals();
	    return(sig);
	  }
	else
	  {
	    unlockSignals();
	    return(0);
	  }
      }
    else
      return(0);
  }

  size_t
  ALingA::signalCount()
  {
    ALingAImpl * impl = static_cast<ALingAImpl *>(impl_);

    impl->waitForCounts();
    return(impl->signalCount_);
  }

  size_t
  ALingA::lingaCount()
  {
    ALingAImpl * impl = static_cast<ALingAImpl *>(impl_);

    impl->waitForCounts();
    return(impl->lingaCount_);
  }

  size_t
  ALingA::countSignals() const
  {
    ALingAImpl * impl = static_cast<ALingAImpl *>(impl_);

    size_t count = 0;
    impl->lockSignals();
    count = impl->signals_.size();
    impl->unlockSignals();
    // gives number of associated signals
    return(count);
  }

  size_t
  ALingA::countLingAs() const
  {
    ALingAImpl * impl = static_cast<ALingAImpl *>(impl_);


    // gives number of associated lingas as of time of calling
    size_t count = 0;
    impl->lockLingas();
    count = impl->lingas_.size();
    impl->unlockLingas();
    return(count);
  }
  
  Signal &
  ALingA::signal(
		 long	track
		 ) const
  {
    ALingAImpl * impl = static_cast<ALingAImpl *>(impl_);

    impl->lockSignals();
    ALingAImpl::iSigs isigs = impl->signals_.find(track);
    if ( isigs != impl->signals_.end() )
      {
	impl->unlockSignals();
	return(*isigs->second);
      }
    impl->unlockSignals();
    throw NoSignalAtTrack(track);
  }

  void
  ALingAImpl::read()
  {
    // read ALingA headers
    readHeaders();

    // signals_ are unlocked within this wait
    waitForSignalsDone();
    // Done with this ALingA bitstream

    delete reader_;
  }

  void
  ALingAImpl::readHeaders()
  {
    DriverParser driver(*this);
    reader_ = &intf_->reader();
    Ogg::Logical::Reader & r = *reader_;

    // skip header 0 Ogg-format preamble already checked in selectCallBack
    size_t	off = 12;
    driver.parse(reinterpret_cast<char *>(r->data())+off
		 ,r->size()-off
		 );

    while( !doneParsing_ )
      {
	++r;
	driver.parse(r->data()
		     ,r->size()
		     );
      }
  }

  void
  DriverParser::on_start_element(
				 const Glib::ustring& 	name
				 , const xmlpp::SaxParser::AttributeList& attributes
				 )
  {
    if ( name == "ALingA" )
      new ALingAXMLparser(top()		// this
			  ,alingaImpl_
			  ,attributes
			  );
    else
      throw XMLparser::ContextCorruption("ALingA"
					 ,name
					 );
  }
  
  ALingAXMLparser::ALingAXMLparser(
				   XMLparser &			driverParser
				   , ALingAImpl &		alingaImpl
				   , const xmlpp::SaxParser::AttributeList& attributes
				   )
    : XMLparser("ALingA"
		,&driverParser
		)
    , alingaImpl_(alingaImpl)
  {
    for ( xmlpp::SaxParser::AttributeList::const_iterator i = attributes.begin()
	    ;
	  i != attributes.end()
	    ;
	  ++i
	  )
      {
	
	if ( i->name == "xmlns" )
	  {
	    alingaImpl_.xmlns_ = i->value;
	  }
	else if ( i->name == "signalCount" )
	  {
	    long l = atol(i->value.c_str());
	    if ( l <= 0 )
	      // signalCount is not essential
	      // rely on actual Signal elements
	      l = 1;
	    alingaImpl_.signalCount_ = l;
	  }
	else if ( i->name == "lingaCount" )
	  {
	    long l = atol(i->value.c_str());
	    if ( l <= 0 )
	      // lingaCount IS essential
	      // but it just means not all the lingas will be read
	      l = 1;
	    alingaImpl_.lingaCount_ = l;
	  }
	else if (
		 // already checked - just let it go
		 ( i->name != "dtdVersion" )
		 &&
		 ( i->name != "xmlns:xlink" )
		 )
	  {
	    alingaImpl_.manifold_(*i);
	  }
      }
    
    alingaImpl_.gotCounts();
  }
    
  void
  ALingAXMLparser::on_start_element(
				    const Glib::ustring& 	name
				    , const xmlpp::SaxParser::AttributeList& attributes
				    )
  {
    if ( name == "Signal" )
      new SignalXMLparser(top()
			  ,alingaImpl_
			  ,attributes
			  );
    else
      throw XMLparser::ContextCorruption("Signal"
					 ,name
					 );
  }

  SignalXMLparser::SignalXMLparser(
				   XMLparser &		alingaXmlParser
				   ,ALingAImpl &	alingaImpl
				   ,const xmlpp::SaxParser::AttributeList& attributes
				   )
    : XMLparser("Signal"
		,&alingaXmlParser
		)
    , alingaImpl_(alingaImpl)
    , signalImpl_(0)
  {
    long	 track = 0;

    for ( xmlpp::SaxParser::AttributeList::const_iterator i = attributes.begin()
	    ;
	  i != attributes.end()
	    ;
	  ++i
	  )
      {
	long long	ll = 0;
	if ( (i->name != "track")
	     &&
	     (i->name != "codecType")
	     &&
	     (i->name != "xlink:type")
	     &&
	     (i->name != "xlink:href")
	     )
	  {
	    ll = atoll(i->value.c_str());
	    if ( ll <= 0 )
	      ll = 0;
	  }

	if ( i->name == "track" )
	  {
	    track = atol(i->value.c_str());
	    if ( track != 0 )
	      {
		signalImpl_ = new SignalImpl(track);
	      }
	    else
	      alingaImpl_.doneParsing_ = true;
	  }
	else if ( i->name ==  "codecType" )
	  {
	    if ( i->value.empty() )
	      signalImpl_->codecType_ = "** Error - no codecType **";
	    else
	      signalImpl_->codecType_ = i->value;
	  }
	else if ( i->name ==  "xlink:href" )
	  {
	    signalImpl_->href_ = i->value;
	  }
	else if ( i->name ==  "baseGransPerFrame" )
	    {
	      signalImpl_->baseGransPerFrame_ = size_t(ll);
	    }
	else if ( i->name ==  "timeBias" )
	  {
	    signalImpl_->bias_.time = Position(ll);
	  }
	else if ( i->name ==  "spaceBias" )
	  {
	    signalImpl_->bias_.space = Position(ll);
	  }
	else if ( i->name ==  "timeDecimation" )
	  {
	    signalImpl_->decimation_.time = Position(ll);
	  }
	else if ( i->name ==  "spaceDecimation" )
	  {
	    signalImpl_->decimation_.space = Position(ll);
	  }

	if ( !signalImpl_ )
	  {
	    // sink it now
	    break;
	  }
      }

    alingaImpl_.gotSignal(signalImpl_);

  }

  void
  SignalXMLparser::on_start_element(
				    const Glib::ustring & name
				    ,const xmlpp::SaxParser::AttributeList& attributes
				    )
  {
    if ( name == "Metas" )
      {
	new MetasXMLparser(top()
			   ,signalImpl_->metas_
			   );
      }
    else
      throw XMLparser::ContextCorruption("Metas"
					 ,name
					 );
  }

}
