/*===========================================================================*/
/*
 * This file is part of
 * libalinga - a C++ library for Alligned Linguistic Annotation
 *
 * Copyright (C) 2007, 2008 Elaine Tsiang YueLien
 *
 * libalinga is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, see http://www.gnu.org/licenses, or write to the
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301, USA
 *
 *===========================================================================*/
/** @file ALingAImpl.H                                                       *
 *                                                                           *
 *  ALingA::ALingAImpl - ALingA implementation interface                     *
 *                                                                           */
/*===========================================================================*/
#ifndef	ALingAImplementationInterface
#define	ALingAImplementationInterface

#ifdef __GNUG__
#pragma interface
#endif

#include	<config.h>
#include	<ALingA/Encode.H>
#include	<ALingA/XMLwriter.H>
#include	<ALingA/XMLparser.H>
#include	<ALingA/ALingA.H>

namespace	ALingA
{
  class	ALingAImpl;
  class	DriverXMLparser;
  class ALingAXMLparser;
  class SignalXMLparser;
}

#include	<SignalImpl.H>

#include	<Ogg/Logical.H>


namespace	ALingA
{
  class	DriverXMLparser;
  class ALingAXMLparser;
  class SignalXMLparser;

  class ALingAImpl
  {
    friend class ALingA;
    friend class DriverXMLparser;
    friend class ALingAXMLparser;
    friend class SignalXMLparser;

  private:
    static
    const unsigned short	dtdMajor_ = ALINGA_DTD_MAJOR;
    static
    const unsigned short	dtdMinor_ = ALINGA_DTD_MINOR;

  private:
    ALingA *			intf_;

    bool			writing_;
    Glib::ustring		xmlns_;
    Manifold			manifold_;

    pthread_cond_t		condCount_;
    pthread_mutex_t		lockCount_;
    size_t			signalCount_;	// parsed signal count
    size_t			lingaCount_;	// parsed linga count

    bool			doneParsing_;
    Ogg::Logical::Reader *	reader_;
 
    pthread_cond_t				condSignals_;
    pthread_mutex_t				lockSignals_;
    map<long, SignalImpl*>			signals_;
    typedef map<long, SignalImpl *>::iterator	iSigs;
    size_t					argSigs_;

    pthread_mutex_t				lockLingas_;
    map<long/*track*/, LingA *>			lingas_;
    typedef map<long, LingA *>::iterator	iLings;

    ALingAImpl(
	       bool	writing
	       );

    ~ALingAImpl()
    {}

    Signal &
    add(
	const long		track
	,const Manifold &	manifold
	);

    void
    remove(
	   const long		track
	   );

    void
    add(
	LingA &
	);

    void
    remove(
	LingA &
	);

    void
    write();

    void
    writeHeaders();

    void
    read();

    void
    readHeaders();

    Signal *
    signal(); 

    void
    lockSignals()
    {
      pthread_mutex_lock(&lockSignals_);
    }
	
    void
    unlockSignals()
    {
      pthread_mutex_unlock(&lockSignals_);
    }
	
    void
    lockLingas()
    {
      pthread_mutex_lock(&lockLingas_);
    }
	
    void
    unlockLingas()
    {
      pthread_mutex_unlock(&lockLingas_);
    }

    void
    waitForCounts();

    void
    gotCounts();

    void
    waitForSignal();

    void
    gotSignal(SignalImpl *);

    void
    waitForSignalsReady();

    void
    waitForSignalsDone();

    void
    reset();
 };

  class DriverParser : public XMLparser
  {
    friend class	ALingAImpl;
    friend class	XMLparser;

  private:
    ALingAImpl &		alingaImpl_;

    DriverParser(
		 ALingAImpl &	alingaImpl
		 )
      : XMLparser("Driver")
      , alingaImpl_(alingaImpl)
    {}

    ~DriverParser()
    {}

    /* base behaviour
       void
       on_start_document();

       void
       on_end_document();
    */

    void
    on_start_element(
		     const Glib::ustring& 	name
		     , const xmlpp::SaxParser::AttributeList& attributes
		     );
    
    void
    on_end_element(
		   const Glib::ustring& 	name
		   )
    {
      throw XMLparser::UnexpectedElement(name);
    }
    
    void
    on_characters(
		  const Glib::ustring& characters
		  )
    {
      if ( characters != "\n" )
	throw XMLparser::UnexpectedCharacters(characters);

    }
    
    void on_cdata_block(
			const Glib::ustring& text
			)
    {
      throw XMLparser::UnexpectedCData(text);
    }
    
  };

  class ALingAXMLparser : public XMLparser
  {
    friend class	ALingAImpl;
    friend class	XMLparser;

  private:
    ALingAImpl &		alingaImpl_;

  public:
    ALingAXMLparser(
		    XMLparser &			xmlParser
		    , ALingAImpl &		alingaImpl
		    , const xmlpp::SaxParser::AttributeList& attributes
		    )
      ;

    ~ALingAXMLparser()
    {}


    void
    on_start_document()
    {
      throw XMLparser::UnexpectedDocument();
    }
    

    void
    on_end_document()
    {
      throw XMLparser::UnexpectedDocument();
    }
    


    void
    on_start_element(
		     const Glib::ustring& 	name
		     , const xmlpp::SaxParser::AttributeList& attributes
		     );
    /* base behaviour
       void
       on_end_element(
       const Glib::ustring& 	name
       );
    */
    
    void
    on_characters(
		  const Glib::ustring& characters
		  )
    {
      if ( characters != "\n" )
	throw XMLparser::UnexpectedCharacters(characters);
    }
    
    void on_cdata_block(
			const Glib::ustring& text
			)
    {
      throw XMLparser::UnexpectedCData(text);
    }
  };

  class SignalXMLparser : public XMLparser
  {
  private:
    ALingAImpl &		alingaImpl_;
    SignalImpl *		signalImpl_;

  public:
    SignalXMLparser(
		    XMLparser &		alingaXmlParser
		    ,ALingAImpl &		alingaImpl
		    , const xmlpp::SaxParser::AttributeList& attributes
		    )
      ;

    ~SignalXMLparser()
    {}


    void
    on_start_document()
    {
      throw XMLparser::UnexpectedDocument();
    }
    

    void
    on_end_document()
    {
      throw XMLparser::UnexpectedDocument();
    }
    


    void
    on_start_element(
		     const Glib::ustring& 	name
		     , const xmlpp::SaxParser::AttributeList& attributes
		     );

    /* base behaviour
       void
       on_end_element(
       const Glib::ustring& 	name
       );
    */
    
    void
    on_characters(
		  const Glib::ustring& characters
		  )
    {
      if ( characters != "\n" )
	throw XMLparser::UnexpectedCharacters(characters);
    }
    
    void on_cdata_block(
			const Glib::ustring& text
			)
    {
      throw XMLparser::UnexpectedCData(text);
    }
  };
}

#endif
