/*===========================================================================*/
/*
 * This file is part of
 * libalinga - a C++ library for Alligned Linguistic Annotation
 *
 * Copyright (C) 2007, 2008 Elaine Tsiang YueLien
 *
 * libalinga is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, see http://www.gnu.org/licenses, or write to the
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301, USA
 *
 *===========================================================================*/
/** @file LingAImpl.C                                                        *
 *                                                                           *
 *  ALingA::LingA implementation                                             *
 *                                                                           */
/*===========================================================================*/
#ifdef __GNUG__
#pragma implementation "LingA.H"
#pragma implementation "LingAImpl.H"
#endif

#include	<ALingA/Encode.H>
#include	<LingAImpl.H>
#include	<Ogg/Debug.H>

#include	<sstream>
#include	<limits>

namespace ALingA
{
  using namespace	Ogg;

  /*------ Writing LingA ----------------------------------------------------------*/

  LingA::LingA(
	       ALingA &		alinga
	       ,long		track
	       ,long		signalTrack
	       )
    : Logical(alinga.transport()
	      ,track
	      )
  {
    LingAImpl *impl = new LingAImpl(
				    alinga
				    ,signalTrack
				    );

    impl->intf_ = this;
    impl_ = impl;

    alinga.add(*this);
  }

  LingAImpl::LingAImpl(
		       ALingA &		alinga
		       ,long		signalTrack
		       )
    : alinga_(alinga)
    , signalTrack_(signalTrack)
    , signal_(0)
    , metas_()
    , writer_(0)
    , xmlWriter_(0)
    , lingaWriter_(0)
    , lastNodeWritten_(0)
    , xmlParser_(0)
    , reader_(0)
    , nodesById_()
    , nodesByTime_()
    , arcsById_()
    , arcsByEpoch_()
    , arcsByLabel_()
    , doneParsing_(false)
  {}

  LingA::~LingA()
  {
    if ( impl_ )
      {
	LingAImpl * impl = static_cast<LingAImpl *>(impl_);
	impl->alinga_.remove(*this);
	delete impl;
	impl_ = 0;
      }
  }

  ALingA &
  LingA::alinga() const
  {
    LingAImpl * impl = static_cast<LingAImpl *>(impl_);

    return(impl->alinga_);
  }

  Metas &
  LingA::metas()
  {
    LingAImpl * impl = static_cast<LingAImpl *>(impl_);
    
    return(impl->metas_);
  }

  Signal const &
  LingA::signal()
  {
    LingAImpl * impl = static_cast<LingAImpl *>(impl_);
    
    if ( !impl->signal_ )
      impl->signal_ = &impl->alinga_.signal(impl->signalTrack_);

    return(*impl->signal_);
  }

  void
  LingAImpl::write()
  {
    // Let Ogg take care of duplicate writes

    // LingA already knows which signal track it is based on, whether interleaved.
    // So here it will call the ALingA to get the current Signal for that track.
    // It could bomb out on ALingA::NoSignalAtTrack, which it should.
    signal_ = & alinga_.signal(signalTrack_);

    writer_ = &intf_->writer();
    lastNodeWritten_ = 0;

    Ogg::Logical::Writer & writer = *writer_;

    xmlWriter_ = new XMLwriter(
			       "LingA"
			       ,dtdMajor_
			       ,dtdMinor_
			       );

    // throw away the preamble by writing it
    // and then overwriting it
    xmlWriter_->write(writer->data());

    char * const	data = reinterpret_cast<char *>(writer->data());
    size_t		off = 12;

    // Ogg-format preamble in header 0
    data[0] = 'L';data[1]='i';data[2]='n';data[3]='g';data[4]='A';
    data[5] = 0;data[6]=0;data[7]=0;

    // Encode will impose big-endian order
    Encode		enc(Unsigned16Bit);
    unsigned short *	u = reinterpret_cast<unsigned short *>(data+8);
    *u = dtdMajor_;
    enc.encode(u
	       ,sizeof(unsigned short)
	       );
    ++u;
    *u = dtdMinor_;
    enc.encode(u
	       ,sizeof(unsigned short)
	       );


    lingaWriter_ = new XMLwriter(
				 *xmlWriter_
				 ,"LingA"
				 );
    {
      LongAttribute	trk("track"
			    ,intf_->serialNo()
			    );
      lingaWriter_->attribute(
			      trk.name
			      ,trk.value
			      );
    }
    {
      LongAttribute	trk("alingaTrack"
			    ,alinga_.track()
			    );
      lingaWriter_->attribute(
			      trk.name
			      ,trk.value
			      );
    }
    {
      LongAttribute	trk("signalTrack"
			    ,signalTrack_
			    );
      lingaWriter_->attribute(
			      trk.name
			      ,trk.value
			      );
    }

    lingaWriter_->endAttributes();

    if ( !metas_.empty() )
      {
	XMLwriter metasWriter(
			      *lingaWriter_
			      ,"Metas"
			      );
	metasWriter.endAttributes();
	metasWriter.metas(metas_);
      }
    // metas tag closed

    writer.dataSize(off+xmlWriter_->textSize());
    xmlWriter_->write(
		      reinterpret_cast<char *>(writer->data())+off
		      );

    // Done writing LingA header
    ++writer;
    writer.beginData();
  }

  ArcImpl &
  LingAImpl::arcImpl(iArcs & iarc) const
  {
    Arc * arc = *iarc;

    return(*static_cast<ArcImpl *>(arc->impl()));
  }

  void
  LingAImpl::write(
		   XMLwriter &	nodeWriter
		   ,ArcImpl &	arc
		   )
  {
    XMLwriter arcWriter(nodeWriter
			 ,"Arc"
			 );
    {
      IdAttribute ia("Id"
		     ,arc.id_
		     );
      arcWriter.attribute(
			  ia.name
			  ,ia.value
			  );
    }
    {
      IdAttribute ia("leftNode"
		     ,arc.left_->id_
		     );
      arcWriter.attribute(
			  ia.name
			  ,ia.value
			  );
    }
    {
      IdAttribute ia("rightNode"
		     ,arc.right_->id_
		     );
      arcWriter.attribute(
			  ia.name
			  ,ia.value
			  );
    }
    arcWriter.endAttributes();

    if ( arc.metas_ 
	 &&
	 !arc.metas_->empty() )
      {
	XMLwriter metasWriter(
			      arcWriter
			      ,"Metas"
			      );
	metasWriter.endAttributes();
	metasWriter.metas(*arc.metas_);
      }
    // metas tag closed
    {
      XMLwriter labelWriter(
			    arcWriter
			    ,"Label"
			    );
      labelWriter.attribute(
			    "randomVar"
			    ,arc.label_.randomVar
			    );
      if ( !arc.label_.likelihood.empty() )
	{
	  labelWriter.attribute(
				"likelihood"
				,arc.label_.likelihood
				);
	}
      labelWriter.endAttributes();

      labelWriter.characters(
			     arc.label_.value
			     );
    }
    // Label tag closed

    // Arc tag will close upon exit
  }

  Position
  LingAImpl::basePosition(NodeImpl &	node)
  {
    long long time = node.time_;
    long long space = node.space_;

    if ( time < 0 )
      {
	time = node.id_;
      }

    if ( space < 0 )
      {
	space = 0;
      }

    return(signal_->basePosition(Position(time)
				 , Position(space)
				 )
	   );
  }

  void
  LingAImpl::write(
		   NodeImpl &	node
		   )
  {
    XMLwriter nodeWriter(*lingaWriter_
			 ,"Node"
			 );
    {
      IdAttribute ia("Id"
		     ,node.id_
		     );
      nodeWriter.attribute(
			   ia.name
			   ,ia.value
			   );
    }
    Glib::ustring idrefs;
    {
      for ( Arcs::iterator iarcs=node.rights_.begin()
	      ;
	    iarcs != node.rights_.end()
	      ;
	    ++iarcs
	    )
	{
	  IdAttribute ia("x"
			 ,arcImpl(iarcs).id_
			 );
	  idrefs += " ";
	  idrefs += ia.value;
	}
    }
    nodeWriter.attribute(
			 "rightArcs"
			 ,idrefs
			 );
    if ( 0 <= node.time_ )
      {
	LongLongAttribute lla("time"
			      ,node.time_
			      );
	nodeWriter.attribute(
			     lla.name
			     ,lla.value
			     );
      }
    if ( 0 <= node.space_ )
      {
	LongLongAttribute lla("space"
			      ,node.space_
			      );
	nodeWriter.attribute(
			     lla.name
			     ,lla.value
			     );
      }

    nodeWriter.endAttributes();

    for ( Arcs::iterator iarcs=node.lefts_.begin()
	    ;
	  iarcs != node.lefts_.end()
	    ;
	  ++iarcs
	  )
      {
	write(nodeWriter
	      ,arcImpl(iarcs)
	      );
      }
  }

  void
  LingA::write(
	       Position		time
	       )
  {
    static_cast<LingAImpl *>(impl_)->write(time);    
  }

  void
  LingAImpl::write(
		   Position	time
		   )
  {
    iNodesByTime	inodes;
    iNodesByTime	first;

    if ( !writer_ )
      write();

    if ( !lastNodeWritten_ )
      {
	if ( nodesByTime_.empty() )
	  return;
	else
	  first = nodesByTime_.begin();
      }
    else
      {
	first = nodesByTime_.lower_bound(lastNodeWritten_->time_);
      }


    for (inodes = first
	 ;
	 (
	  (inodes->first <= time)
	  &&
	  (inodes != nodesByTime_.end())
	  )
	   ;
	 ++inodes
	 )
      {
	if ( !inodes->second->written_
	     &&
	     !inodes->second->detached()	// don't write detached nodes
	     )
	  {
	    write(*inodes->second);
	    writer_->dataSize(xmlWriter_->textSize());
	    xmlWriter_->write(
			      reinterpret_cast<char *>((*writer_)->data())
			      );
	    writer_->granulePosition(basePosition(*(inodes->second)));
	    ++(*writer_);
	    Ogg::debug(true
		       ,"Serial ", intf_->serialNo()
		       ," wrote node ", inodes->second->id_
		       ," at time ", inodes->first
		       );
	    writer_->flush();
	    inodes->second->written_ = true;
	    lastNodeWritten_ = inodes->second;
	  }
      }
    for ( inodes = first
	 ;
	 (
	  (inodes->first <= time)
	  &&
	  (inodes != nodesByTime_.end())
	  )
	   ;
	 ++inodes
	 )
      {
	if ( inodes->second->detached() )
	  {
	    delete inodes->second->intf_;
	  }
      }
  }

  void
  LingA::end()
  {
    static_cast<LingAImpl *>(impl_)->end();
  }

  void
  LingAImpl::end()
  {
    // Close LingA tag
    delete lingaWriter_;
    lingaWriter_ = 0;

    writer_->dataSize(xmlWriter_->textSize());
    xmlWriter_->write(
		      reinterpret_cast<char *>((*writer_)->data())
		      );
    // Write out closing LingA tag
    delete writer_;
    writer_ = 0; 

    delete xmlWriter_;
    xmlWriter_ = 0;
 }

  Id
  LingAImpl::nextNodeId()
  {
    Id	id = nodesById_.size() + 1;

    while ( nodesById_.find(id) != nodesById_.end() )
      {
	++id;
      }
    return(id);
  }

  Id
  LingAImpl::nextArcId()
  {
    Id	id = arcsById_.size() + 1;

    while ( arcsById_.find(id) != arcsById_.end() )
      {
	++id;
      }
    return(id);
  }

  void
  LingAImpl::node(NodeImpl & n)
  {
    nodesById_[n.id_] = &n;
    if ( 0 <= n.time_ )
      nodesByTime_.insert(std::make_pair(Position(n.time_)
					 ,&n
					 )
			  );
  }

  void
  LingAImpl::erase(NodeImpl & n)
  {
    iNodesById i = nodesById_.find(n.id_);
    if ( i != nodesById_.end() )
      nodesById_.erase(i);

    if ( 0 <= n.time_ )
      {
	std::pair<iNodesByTime, iNodesByTime> r
	  = nodesByTime_.equal_range(Position(n.time_));
	for ( iNodesByTime it = r.first
		;
	      it != r.second
		;
	      ++it
	    )
	  {
	    if ( it->second->id_ == n.id_ )
	      {
		nodesByTime_.erase(it);
		break;
	      }
	  }
	// If this node is not in nodesByTime_, there was an error
	// but doesn't matter now.
      }
  }

  void
  LingAImpl::arc(ArcImpl & a)
  {
    arcsById_[a.id_] = &a;
    
    if ( a.epochDefined() )
      arcsByEpoch_.insert(std::make_pair(Epoch(a.left_->time_
					       ,a.right_->time_
					       )
					 , &a
					 )
			  );
    arcsByLabel_.insert(std::make_pair(a.label_
				       , &a
				       )
			);
  }

  void
  LingAImpl::erase(ArcImpl & a)
  {
    iArcsById i = arcsById_.find(a.id_);
    if ( i != arcsById_.end() )
      arcsById_.erase(i);

    if ( a.epochDefined() )
      {
	std::pair<iArcsByEpoch, iArcsByEpoch> r = arcsByEpoch_.equal_range(a.intf_->epoch());
	for ( iArcsByEpoch it = r.first
		;
	      it != r.second
		;
	      ++it
	    )
	  {
	    if ( it->second->id_ == a.id_ )
	      {
		arcsByEpoch_.erase(it);
		break;
	      }
	  }
	// If this arc is not in arcsByEpoch_, there was an error
	// but doesn't matter now.
      }

    std::pair<iArcsByLabel, iArcsByLabel> r = arcsByLabel_.equal_range(a.label_);
    for ( iArcsByLabel it = r.first
	    ;
	  it != r.second
	    ;
	  ++it
	  )
      {
	if ( it->second->id_ == a.id_ )
	  {
	    arcsByLabel_.erase(it);
	    break;
	  }
      }
  }

  /*------ Reading LingA ----------------------------------------------------------*/
  
  LingA::LingA(
	       ALingA &		alinga
	       )
    : Logical(alinga.transport())
  {
    LingAImpl *impl = new LingAImpl(
				    alinga
				    );

    impl->intf_ = this;
    impl_ = impl;
    // can't alinga.add(*this) because no serialNo yet
  }

  LingAImpl::LingAImpl(
		       ALingA &		alinga
		       )
    : alinga_(alinga)
    , signalTrack_(0)
    , signal_(0)
    , metas_()
    , writer_(0)
    , xmlWriter_(0)
    , lingaWriter_(0)
    , lastNodeWritten_(0)
    , reader_(0)
    , nodesById_()
    , nodesByTime_()
    , arcsById_()
    , arcsByEpoch_()
    , arcsByLabel_()
    , doneParsing_(false)
  {
  }

  bool
  LingA::selectCallback(
			Ogg::Packet &	pkt
			)
  {
    char * header = static_cast<char *>(pkt.data());

    if( 
       !strncmp(header, "LingA", 5)
       &&
       ( header[5] == '\0' )
       &&
       ( header[6] == '\0' )
       &&
       ( header[7] == '\0' )
	)
      {
	Encode			enc(Unsigned16Bit);
	unsigned short *	uMajor = reinterpret_cast<unsigned short *>(header+8);
	unsigned short *	uMinor = reinterpret_cast<unsigned short *>(header+10);
	enc.decode(uMajor
		   ,sizeof(unsigned short)
		   );
	enc.decode(uMinor
		   ,sizeof(unsigned short)
		   );
	if ( (*uMajor != LingAImpl::dtdMajor_)
	     ||
	     (*uMinor != LingAImpl::dtdMinor_)
	     )
	  throw ALingA::InconsistentDTDversion(
					       *uMajor
					       ,*uMinor
					       ,LingAImpl::dtdMajor_
					       ,LingAImpl::dtdMinor_
					       );

	return(true);
      }
    return(false);
  }

  void
  LingAImpl::readHeader()
  {
    // parse header packets
    Ogg::Logical::Reader & reader = *reader_;
    xmlParser_ = new LingADriverParser(
				       *this
				       );

    // skip header 0 Ogg-format preamble already checked in selectCallBack
    size_t	off = 12;
    xmlParser_->parse(reinterpret_cast<char *>(reader->data())+off
		      ,reader->size()-off
		      );
    ++reader;
  }

  void
  LingAImpl::readNode()
  {
    if ( !reader_ )
      {
	// reading first node
	reader_ = &intf_->reader();
	doneParsing_ = false;

	readHeader();
      }

    Ogg::Logical::Reader & reader = *reader_;

    if ( !doneParsing_ )
      xmlParser_->parse(reinterpret_cast<char *>(reader->data())
			,reader->size()
			);
    if ( !reader.ending() )
      ++reader;
  }

  void
  LingAImpl::erase()
  {
    nodesByTime_.clear();
    for ( iNodesById i = nodesById_.begin()
	    ;
	  i != nodesById_.end()
	    ;
	  ++i
	  )
      {
	delete i->second;
      }
    nodesById_.clear();

    arcsByEpoch_.clear();
    arcsByLabel_.clear();
    for ( iArcsById i = arcsById_.begin()
	    ;
	  i != arcsById_.end()
	    ;
	  ++i
	  )
      {
	delete i->second;
      }
    arcsById_.clear();

    signalTrack_ = 0;
    signal_ = 0;
    delete reader_;
    reader_ = 0;
    doneParsing_ = false;
  }

  bool
  LingAImpl::operator==(LingAImpl & la2)
  {
    return( this == &la2 );
  }

  void
  LingA::node(
	      Arc & //don't care
	      )
  {
    LingAImpl * impl = static_cast<LingAImpl *>(impl_);
 
    if ( !impl->reader_ )
      {
	impl->readNode();

	// now this LingA has a serialNo
	impl->alinga_.add(*this);
      }
    else
      impl->readNode();
  }

  void
  LingA::erase()
  {
    LingAImpl * impl = static_cast<LingAImpl *>(impl_);

    impl->erase();
  }

  LingA &
  LingA::operator+=(LingA & linga)
  {
    Epoch epoch(0, std::numeric_limits<long>::max());
    Nodes nodes = linga.nodesIn(epoch);
    std::map<Id, Node *> newNodes;

    debug(true,
	  "in LingA::operator+= to copy ", nodes.size(), " nodes"
	  );
    for ( iNodes i = nodes.begin()
	    ;
	  i != nodes.end()
	    ;
	  ++i
	  )
      {
	Node & node = **i;
	Node & newNode = *new Node(*this, node.time());
	// key is OLD id, NOT new id
	newNodes[node.id()] = &newNode;

	Arcs leftArcs = node.lefts();
	for ( iArcs ii = leftArcs.begin()
		;
	      ii != leftArcs.end()
		;
	      ++ii
	      )
	  {
	    // we retrieve the left node by the OLD id
	    std::map<Id, Node*>::iterator inew = newNodes.find((*ii)->left().id());
	    if ( inew != newNodes.end() )
	      {
		Arc  & arc = *new Arc(*(inew->second)
				   , newNode
				   , (*ii)->label()
				   );
		debug(true,
		      "augmented with ", arc.label().randomVar, " of  id ", arc.id(), ", ",
		      arc.label().value, "[", arc.left().time(), ",", arc.right().time(), "]"
		      );
		if ( !(*ii)->metas().empty() )
		  {
		    arc.metas() = (*ii)->metas();
		  }
	      }
	    else
	      {
		debug(true,
		      "left node of id ", inew->second->id(), " has not already been received!"
		      );
	      }
	  }
      }

    return(*this);
  }

  /*------ Searching LingA ----------------------------------------------------------*/
  Arcs
  LingA::arcsCrossing(Epoch & epoch)
  {
    return(static_cast<LingAImpl *>(impl_)->arcsCrossing(epoch, true));    
  }

  Arcs
  LingA::arcsAt(Epoch & epoch)
  {
    return(static_cast<LingAImpl *>(impl_)->arcsCrossing(epoch, false));    
  }

  Arcs
  LingAImpl::arcsCrossing(
			  Epoch & epoch
			  , bool includeContaining
			  )
  {
    Arcs arcs = Arcs();
    std::pair<iArcsByEpoch, iArcsByEpoch> r = arcsByEpoch_.equal_range(epoch);
    EpochRelation er;
    EpochRelation::Type ert;
    // find all arcs CONTAINED in or LEFT of epoch
    iArcsByEpoch it = r.first;
    while ( it != arcsByEpoch_.begin() )
      {
	--it;
	ert = er(it->second->intf_->epoch()
		,epoch);
	if  ( ert <= EpochRelation::BEFORE )
	  {
	    break;
	  }
	Ogg::debug(false
		   ,"LingAImpl::arcsCrossing: ["
		   ,epoch.left,",",epoch.right,"]"
		   ,it->second->intf_->label().value
		   ,"[",it->second->intf_->epoch().left,",",it->second->intf_->epoch().right,"]"
		   );	
	r.first = it;
      }
    it = r.second;
    // find all arcs CONTAINING or RIGHT of epoch
    while ( it != arcsByEpoch_.end() )
      {
	ert = er(it->second->intf_->epoch()
		,epoch);
	if  ( EpochRelation::AFTER <= ert )
	  {
	    break;
	  }
	Ogg::debug(false
		   ,"LingAImpl::arcsCrossing: ["
		   ,epoch.left,",",epoch.right,"], "
		   ,it->second->intf_->label().value
		   ,"[",it->second->intf_->epoch().left,",",it->second->intf_->epoch().right,"]"
		   );	
	++it;
	r.second = it;
      }

    if ( includeContaining ) {
      for ( iArcsByEpoch it = r.first
	      ;
	    (it != r.second)
	      ;
	    ++it
	    )
	{
	  Ogg::debug(false
		     ,"LingAImpl::arcsCrossing: ["
		     ,epoch.left,",",epoch.right,"], gathered "
		     ,it->second->intf_->label().value
		     ,"[",it->second->intf_->epoch().left,",",it->second->intf_->epoch().right,"]"
		     );	
	  arcs.insert(it->second->intf_);
	}
    }
    else
      {
      for ( iArcsByEpoch it = r.first
	      ;
	    (it != r.second)
	      ;
	    ++it
	    )
	{
	  ert = er(it->second->intf_->epoch()
		   ,epoch);
	  if  ( EpochRelation::CONTAINING != ert )
	    arcs.insert(it->second->intf_);
	}
      }
    return(arcs);
  }

  Nodes
  LingA::nodesIn(Epoch & epoch)
  {
    return(static_cast<LingAImpl *>(impl_)->nodesIn(epoch));    
  }

  Nodes
  LingAImpl::nodesIn(Epoch & epoch)
  {
    Nodes nodes = Nodes(); 
    Ogg::debug(false
	       ,"Serial ", intf_->serialNo()
	       ," have ", nodesById_.size(), " nodes by Id, "
	       ," have ", nodesByTime_.size(), " nodes by time, "
	       ," looking for lower_bound at ", epoch.left
	       );
    // l is equal to or just beyond epoch.left, OK
    iNodesByTime l = nodesByTime_.lower_bound(epoch.left);
    // but not if it is beyond epoch.right or at end
    if ( ( l != nodesByTime_.end() )
	 &&
	 ( l->second->time_ <= epoch.right )
	 )
      {
	Ogg::debug(false
		   ,"Serial ", intf_->serialNo()
		   ," looking for upper_bound at ", epoch.right
		   );
	// u is just beyond epoch.right, cool
	iNodesByTime u = nodesByTime_.upper_bound(epoch.right);
	  
	// because l has been checked to be legit,
	// this loop is guarateed to insert at least one node
	for ( iNodesByTime it = l
		;
	      ( it != u )
		;
	      ++it
	      )
	  {
	    nodes.insert(it->second->intf_);
	    Ogg::debug(false
		       ,"Serial ", intf_->serialNo()
		       , ", found node with id = ", it->second->id_
		       );
	  }
      }
    Ogg::debug(false
	       ,"Serial ", intf_->serialNo()
	       ,", returning ", nodes.size(), " nodes"
	       );
    return(nodes);
  }

  Arcs &
  join(Arcs & sequence1
       ,Arcs & sequence2
       )
  {
    if ( !sequence1.empty() && !sequence2.empty() )
      {
	Arcs * smallerSequence;
	Arcs * largerSequence;
	if ( sequence2.size() > sequence1.size() )
	  {
	    largerSequence = &sequence2;
	    smallerSequence = &sequence1;
	  }
	else
	  {
	    largerSequence = &sequence1;
	    smallerSequence = &sequence2;
	  }

	for ( iArcs ia = smallerSequence->begin() 
		;
	      ia != smallerSequence->end()
		;
	      ++ia
	      )
	  {
	    largerSequence->insert(*ia);
	  }
	return(*largerSequence);
      }
    else if ( !sequence1.empty() )
      return(sequence1);
    else
      return(sequence2);
  }

  Arcs
  LingA::sequenceIn(Epoch & epoch, Arc * arc)
  {
    return(static_cast<LingAImpl *>(impl_)->sequenceIn(epoch, arc));    
  }

  Arcs
  LingAImpl::sequenceIn(Epoch & epoch, Arc * anyLastArc)
  {
    Arcs thisSequence;
    Arcs rightArcs;

    if ( anyLastArc != 0 ) {
      thisSequence.insert(anyLastArc);
      rightArcs = anyLastArc->right().rights();
    }

    if ( rightArcs.empty() )
      {
	Nodes nodes = nodesIn(epoch);
	
	if ( nodes.empty() )
	  {
	    // reached end
	    // return empty container to be filled by caller
	    return(Arcs());
	  }
	// this need to be done here in case rightArcs is empty
	epoch.left = (*nodes.begin())->time()+1;
	rightArcs = (*nodes.begin())->rights();
      }

    while ( ! rightArcs.empty() )
      {
	Arc * arc = *rightArcs.begin();
	// if we have a gap, let's show it
	// set the next epoch to beyond thisSequence
	// otherwise, we could get into an infinite loop
	// because of possibly epoch.left == epoch.right
	epoch.left = arc->right().time()+1;
	if ( arc->epoch().right > epoch.right )
	  {
	    // if the smallest arc doesn't fit inside this epoch
	    // the larger ones won't
	    break;
	  }
	thisSequence.insert(arc);
	Ogg::debug(false
		   ,"LingAImpl::sequenceIn: [",arc->left().time(),",",arc->right().time(),"] "
		   ,arc->label().randomVar," = ", arc->label().value
		   );
	rightArcs = arc->right().rights();
      }

    Ogg::debug(false
	       ,"LingAImpl::SequenceIn: ", thisSequence.size(), " arcs "
	       ," before ", epoch.left
	       );

    Arcs laterSequence = sequenceIn(epoch, 0);

    return(join(thisSequence, laterSequence));
    
  }

  Arcs
  LingA::sequenceContainedIn(Epoch & epoch)
  {
    return(static_cast<LingAImpl *>(impl_)->sequenceContainedIn(epoch));    
  }

  Arcs
  LingAImpl::sequenceContainedIn(Epoch & epoch)
  {
    Arcs thisSequence;
    Arcs rightArcs;

    Nodes nodes = nodesIn(epoch);
	
    if ( nodes.empty() )
      {
	// reached end
	// return empty container to be filled by caller
	return(Arcs());
      }

    for ( iNodes i = nodes.begin()
	    ;
	  i != nodes.end()
	    ;
	  ++i
	  )
      {
	epoch.left = (*i)->time()+1;
	if ( (*i)->lefts().empty() )
	  {
	    rightArcs = (*i)->rights();
	    break;
	  }
      }  

    while ( ! rightArcs.empty() )
      {
	Arc * arc = *rightArcs.begin();
	// if we have a gap, let's show it
	// set the next epoch to beyond thisSequence
	// otherwise, we could get into an infinite loop
	// because of possibly epoch.left == epoch.right
	epoch.left = arc->right().time()+1;
	if ( arc->epoch().right > epoch.right )
	  {
	    // if the smallest arc doesn't fit inside this epoch
	    // the larger ones won't
	    // this sequence is not contained
	    thisSequence.clear();
	    break;
	  }
	thisSequence.insert(arc);
	Ogg::debug(false
		   ,"LingAImpl::sequenceContianedIn: [",arc->left().time(),",",arc->right().time(),"] "
		   ,arc->label().randomVar," = ", arc->label().value
		   );
	rightArcs = arc->right().rights();
      }

    Ogg::debug(false
	       ,"LingAImpl::sequenceContainedIn: ", thisSequence.size(), " arcs "
	       ," before ", epoch.left
	       );

    Arcs laterSequence = sequenceContainedIn(epoch);

    return(join(thisSequence, laterSequence));
    
  }

  Arcs
  LingA::includeArcInSequence(
			      Arc & arc
			      ,Arcs & arcs
			      ,Arc * hint
			      )
  {
    return(static_cast<LingAImpl *>(impl_)->includeArcInSequence(arc
								 , arcs
								 , hint
								 )
	   );
  }

  Arcs
  LingAImpl::includeArcInSequence(
				  Arc & arc
				  ,Arcs & arcs
				  ,Arc * hint
				  )
  {
    if ( arcs.empty() )
      {
	Arcs sequence;
	sequence.insert(&arc);
	return sequence;
      }

    iArcs ia = arcs.begin();

    if ( hint ) {
      ia = arcs.find(hint);
   
      if ( ia == arcs.end() ) {
	// no hint at all
	ia = arcs.begin();
      }
      else if ( (*ia)->right().time() > arc.left().time() ) {
	// bad hint
	ia = arcs.begin();
      }
    }

    iArcs iLastOfLeftSequence = arcs.begin();
    iArcs iFirstOfRightSequence = arcs.end();
    for (
	 ;
	 ia != arcs.end()
	   ;
	 ++ia
	 )
      {
	if ( (*ia)->right().time() <= arc.left().time() )
	  {
	    iLastOfLeftSequence = ia;
	    ++iLastOfLeftSequence;
	  }

	if ( arc.right().time() <= (*ia)->left().time() )
	  {
	    iFirstOfRightSequence = ia;
	    break;
	  }
      }
    
    iArcs iLeft = iLastOfLeftSequence;
    Arcs leftSequence;
    while ( iLeft != arcs.begin() )
      {
	--iLeft;
	if ( leftSequence.empty() ) {
	leftSequence = path(**iLeft, arc);
	if ( !leftSequence.empty() )
	  {
	    Ogg::debug(false
		       ,"LingAImpl::includeArcInSequence: left path from arc "
		       ,(*iLeft)->id(), " to arc ", arc.id()
		       );
	  }
	}
	else {
	  // copy over the rest of the left side of the original sequence
	  leftSequence.insert(*iLeft);
	}
      }
    bool noLeftSequence = false;
    if ( leftSequence.empty() )
      noLeftSequence = true;

    iArcs iRight = iFirstOfRightSequence;
    Arcs rightSequence;
    while ( iRight != arcs.end() )
      {
	if ( rightSequence.empty() ) {
	  rightSequence = path(arc, **iRight);
	  if ( !rightSequence.empty() )
	    {
	      Ogg::debug(false
			 ,"LingAImpl::includeArcInSequence: right path from arc "
			 ,arc.id(), " to arc ", (*iRight)->id()
			 );
	    }
	}
	else {
	  // copy over the rest of the right side of the original sequence
	  rightSequence.insert(*iRight);
	}
	++iRight;
      }

    bool noRightSequence = false;
    if ( rightSequence.empty() )
      noRightSequence = true;

    if ( !leftSequence.empty() && !rightSequence.empty() )
      {
	// arc is in both sequences
	rightSequence.erase(rightSequence.begin());
      }

    // need the bool noXSequence because after this
    // either sequence may be altered
    Arcs & sequence = join(leftSequence, rightSequence);
    
    if ( sequence.empty() )
      {
	// at least have the arc we must include
	sequence.insert(&arc);
      }

    iArcs iSequence = sequence.begin();
    Arcs leftArcs;
    if ( noLeftSequence )
      {
	while ( !(leftArcs = (*iSequence)->left().lefts()).empty() )
	  {
	    // extend left
	    sequence.insert(*leftArcs.begin());
	    iSequence = sequence.begin();
	    Ogg::debug(false
		       ,"LingAImpl::includeArcInSequence: extended left by "
		       ,"[",(*iSequence)->left().time(),",",(*iSequence)->right().time(),"] "
		       ,(*iSequence)->label().randomVar," = ", (*iSequence)->label().value
		       );
	  }

	iLeft = iLastOfLeftSequence;
	bool haveGap = false;
	while ( iLeft != arcs.begin() )
	  {
	    --iLeft;
	    Epoch epoch(0,0);
	    if ( !haveGap ) {
	      if ( (epoch.left = (*iLeft)->right().time()+1)
		   <= (epoch.right = (*iSequence)->left().time()-1)
		   )
		{
		  if ( (*iLeft)->right().rights().empty() ) {
		    // have a gap
		    Arcs gapArcs = sequenceContainedIn(epoch);
		    sequence = join(gapArcs, sequence);
		    sequence.insert(*iLeft);
		    Ogg::debug(false
			       ,"LingAImpl::includeArcInSequence: gap sequence from "
			       ,"[",(*iLeft)->left().time(),",",(*iLeft)->right().time(),"] "
			       ,(*iLeft)->label().randomVar," = ", (*iLeft)->label().value
			       );
		    haveGap = true;
		  }
		}
	    }
	    else {
	      // copy over remaining beyond gap
	      sequence.insert(*iLeft);
	    }
	  }
      }

    if ( noRightSequence )
      {
      iSequence = --sequence.end();
      Arcs rightArcs;
      while ( !(rightArcs = (*iSequence)->right().rights()).empty() )
	{
	  // extend right
	  sequence.insert(*rightArcs.begin());
	  iSequence = --sequence.end();
	  Ogg::debug(false
		     ,"LingAImpl::includeArcInSequence: extended right by "
		     ,"[",(*iSequence)->left().time(),",",(*iSequence)->right().time(),"] "
		     ,(*iSequence)->label().randomVar," = ", (*iSequence)->label().value
		     );
	}

      iRight = iFirstOfRightSequence;
      bool haveGap = false;
      while ( iRight != arcs.end() )
	{
	  Epoch epoch(0,0);
	  if ( !haveGap ) {
	  if ( (epoch.right = (*iRight)->left().time()-1)
	       >= (epoch.left = (*(iSequence))->right().time()+1)
	       )
	    {
	      if ( (*iRight)->left().lefts().empty() ) {
		// have a gap
		Arcs gapArcs = sequenceContainedIn(epoch);
		sequence = join(sequence, gapArcs);
		sequence.insert(*iRight);
		Ogg::debug(false
			   ,"LingAImpl::includeArcInSequence: gap sequence to "
			   ,"[",(*iRight)->left().time(),",",(*iRight)->right().time(),"] "
			   ,(*iRight)->label().randomVar," = ", (*iRight)->label().value
			   );
		haveGap = true;
	      }
	    }
	  }
	  else {
	    // copy over remaining beyond gap
	    sequence.insert(*iRight);
	  }
	  ++iRight;
	}
    }

    return(sequence);
  }    

  /*------ end Searching LingA ------------------------------------------------------*/

  void
  LingADriverParser::on_start_element(
				      const Glib::ustring & name
				      ,const xmlpp::SaxParser::AttributeList& attributes
				      )
  {
    if ( name == "LingA" )
      new LingAXMLparser(*this
			 ,lingaImpl_
			 ,attributes
			 );
    else
      throw XMLparser::ContextCorruption("LingA"
					 ,name
					 );

  }

  LingAXMLparser::LingAXMLparser(
				 LingADriverParser &	driverParser
				 ,LingAImpl &		lingaImpl
				 ,const xmlpp::SaxParser::AttributeList& attributes
				 )
    : XMLparser("LingA"
		,&driverParser
		)
    , lingaImpl_(lingaImpl)
  {
    for ( xmlpp::SaxParser::AttributeList::const_iterator i = attributes.begin()
	    ;
	  i != attributes.end()
	    ;
	  ++i
	  )
      {
	if ( i->name == "signalTrack" )
	  {
	    lingaImpl_.signalTrack_ = atol(i->value.c_str());
	    // whether or not there is a signal at the track
	    // is not relevant on parsing.
	    // When the user wants to know, he'll find out.
	    // So he won't be able to interpret the nodes' Coordinates
	    // with respect to the ALingA manifold.
	    // In any case, he still has all the information from
	    // the LingA stream, and will find out what the error is.
	    // In case of unanchored nodes, this doesn't even matter.
	  }
	/* we don't care about the rest
	if ( i->name == "track" )
	  {
	    long track = atol(i->value.c_str());
	    if ( track != lingaImpl_.serialNo() )
	      throw TrackNotAsAdvertised(lingaImpl_.serialNo()
					 ,track
					 );
	  }
	else if ( i->name == "alingaTrack" )
	  {
	    long track = atol(i->value.c_str());
	    if ( track != lingaImpl_.alinga_.track() )
	      throw TrackNotAsAdvertised(lingaImpl_.alinga_.track()
							    ,track
							    );
	  }
	*/
      }
  }

  void
  LingAXMLparser::on_start_element(
				   const Glib::ustring & name
				   ,const xmlpp::SaxParser::AttributeList& attributes
				   )
  {
    if ( name == "Metas" )
      {
	new MetasXMLparser(top()		// this
			   ,lingaImpl_.metas_
			   );
      }
    else if ( name == "Node" )
      {
	new NodeXMLparser(*this
			  ,lingaImpl_
			  ,attributes
			  );
      }
    else
      throw XMLparser::ContextCorruption("LingA"
					 ,name
					 );

  }

  void
  LingAXMLparser::on_end_element(
				 const Glib::ustring& 	name
				 )
  {
    lingaImpl_.doneParsing_ = true;
  }

  NodeXMLparser::NodeXMLparser(
			       LingAXMLparser &		lingaXMLparser
			       ,LingAImpl &		lingaImpl
			       , const xmlpp::SaxParser::AttributeList& attributes
			       )
    : XMLparser("Node"
		,&lingaXMLparser
		)
    , lingaImpl_(lingaImpl)
  {
    Id	id = 0;
    long long time = -1;
    long long space = -1;

    std::istringstream is;

    for ( xmlpp::SaxParser::AttributeList::const_iterator i = attributes.begin()
	    ;
	  i != attributes.end()
	    ;
	  ++i
	  )
      {
	if ( i->name == "Id" )
	  {
	    long ii = atol(i->value.c_str());
	    if ( 0 < ii )
	      id = Id(ii);
	    debug(true
		  ,"Serial ", lingaImpl_.intf_->track()
		  ," parsing Node with id = ",id
		  );
	  }
	else if ( i->name == "time" )
	  {
	    time = atol(i->value.c_str());
	  }
	else if ( i->name == "space" )
	  {
	    space = atol(i->value.c_str());
	  }
	else if ( i->name == "rightArcs" )
	  {
	    is.str(i->value.raw());
	  }
      }

    iNodesById i;

    if ( id <= 0 )
      {
	throw LingA::NodeHasNoId();
      }
    else if ( (i = lingaImpl_.nodesById_.find(id))
	      == lingaImpl_.nodesById_.end() )
      {
	// make it up now
	if ( lingaImpl_.nodesById_.empty() )
	  {
	    throw LingA::NoNodeOrArcToClone();
	  }
	// clone it
	Node & node = lingaImpl_.nodesById_.begin()->second->intf_->clone(id);
	nodeImpl_ = static_cast<NodeImpl *>(node.impl());
	debug(true
	      ,"Serial ", lingaImpl_.intf_->track()
	      ," cloned Node of id ",nodeImpl_->id_
	      );
      }
    else
      {
	nodeImpl_ = i->second;
	debug(true
	      ,"Serial ", lingaImpl_.intf_->track()
	      ," NOTE: found pre-existing Node of id ",id
	      );
      }
    
    nodeImpl_->time_ = time;
    nodeImpl_->space_ = space;
	
    while ( !is.eof() )
      {
	// expected right arcs
	id = 0;
	long ii = 0;
	is >> ii;
	if ( 0 < ii )
	  {
	    id = Id(ii);

	    ArcImpl * arcImpl = 0;
	    iArcsById ia = lingaImpl_.arcsById_.find(id);
	    if (
		(ia = lingaImpl_.arcsById_.find(id))
		== lingaImpl_.arcsById_.end()
		)
	      {
		// make it up now
		if ( lingaImpl_.arcsById_.empty() )
		  {
		    throw LingA::NoNodeOrArcToClone();
		  }
		// clone it
		// note this is always a clone of the first arc
		Arc & arc = lingaImpl_.arcsById_.begin()->second->intf_->clone(id);
		arcImpl = static_cast<ArcImpl *>(arc.impl());
		debug(true
		      ,"Serial ", lingaImpl_.intf_->track()
		      ," cloned Arc of id ",arcImpl->id_
		      );
	      }
	    else
	      {
		arcImpl = ia->second;
		debug(true
		      ,"Serial ", lingaImpl_.intf_->track()
		      ," NOTE: found pre-existing Arc of id ",id
		      );
	      }
	    // junk until right node shows up
	    arcImpl->junk_ = true;

	    Node * rightNode = 0;
	    i = lingaImpl_.nodesById_.find(std::numeric_limits<Id>::max());
	    if ( i == lingaImpl_.nodesById_.end() )
	      {
		rightNode = &(lingaImpl_.nodesById_.begin()->second->intf_->clone(std::numeric_limits<Id>::max()));
		debug(true
		      ,"Serial ", lingaImpl_.intf_->track()
		      ," cloned right Node of id "
		      ,static_cast<NodeImpl *>(rightNode->impl())->id_
		      );
	      }
	    else
	      {
		rightNode = i->second->intf_;
	      }
	    arcImpl->intf_->right(*rightNode);
	    arcImpl->intf_->left(*nodeImpl_->intf_);
	  }
      }
  }

  void
  NodeXMLparser::on_start_element(
				  const Glib::ustring & name
				  ,const xmlpp::SaxParser::AttributeList& attributes
				  )
  {
    if ( name == "Arc" )
      {
	new ArcXMLparser(*this
			 ,lingaImpl_
			 ,*nodeImpl_
			 ,attributes
			 );
      }
    else
      throw XMLparser::ContextCorruption("LingA"
					 ,name
					 );

  }

  void
  NodeXMLparser::on_end_element(
				 const Glib::ustring& 	name
				 )
  {
    // now enter this node into nodesByTime_
    if ( 0 <= nodeImpl_->time_ )
      lingaImpl_.nodesByTime_.insert(std::make_pair(Position(nodeImpl_->time_)
						    ,nodeImpl_
						    )
				     );
  }

  ArcXMLparser::ArcXMLparser(
			     NodeXMLparser &		nodeXMLparser
			     , LingAImpl &		lingaImpl
			     , NodeImpl &		nodeImpl
			     , const xmlpp::SaxParser::AttributeList& attributes
			     )
    : XMLparser("Arc"
		,&nodeXMLparser
		)
    , lingaImpl_(lingaImpl)
    , nodeImpl_(nodeImpl)
  {
    Id	id = 0;
    Id	leftId = 0;
    Id	rightId = 0;

    for ( xmlpp::SaxParser::AttributeList::const_iterator i = attributes.begin()
	    ;
	  i != attributes.end()
	    ;
	  ++i
	  )
      {
	if ( i->name == "Id" )
	  {
	    long ii = atol(i->value.c_str());
	    if ( 0 < ii )
	      id = Id(ii);
	    debug(true
		  ,"Serial ", lingaImpl_.intf_->track()
		  ," parsing Arc with id = ",id
		  );
	  }
	else if ( i->name == "leftNode" )
	  {
	    leftId = atol(i->value.c_str());
	  }
	else if ( i->name == "rightNode" )
	  {
	    rightId = atol(i->value.c_str());
	  }
      }

    iArcsById ia = lingaImpl_.arcsById_.find(id);
    if ( id <= 0 )
      {
	throw LingA::ArcHasNoId();
      }
    else if ( (ia = lingaImpl_.arcsById_.find(id))
	      == lingaImpl_.arcsById_.end()
	      )
      {
	// make it up now
	if ( lingaImpl_.arcsById_.empty() )
	  {
	    throw LingA::NoNodeOrArcToClone();
	  }
	// clone it
	Arc & arc = lingaImpl_.arcsById_.begin()->second->intf_->clone(id);
	arcImpl_ = static_cast<ArcImpl *>(arc.impl());
	debug(true
	      ,"Serial ", lingaImpl_.intf_->track()
	      ," Note: cloned unannounced Arc of id ",arcImpl_->id_
	      );
      }
    else
      {
	arcImpl_ = ia->second;
	debug(true
	      ,"Serial ", lingaImpl_.intf_->track()
	      ," found pre-existing Arc of id ",id
	      );
      }
    // camel through the needle
    arcImpl_->junk_ = true;

    if ( arcImpl_->intf_->right(*nodeImpl_.intf_) )
      {
	// Later overrides former
	iNodesById i = lingaImpl_.nodesById_.find(leftId);
	if ( i != lingaImpl_.nodesById_.end() )
	  {
	    if ( (
		  // coincidentally
		  (arcImpl_->left_->id_ == leftId)
		  ||
		  // or not
		  arcImpl_->intf_->left(*i->second->intf_)
		  )
		 &&
		 // as advertised
		 ( nodeImpl_.id_ == rightId )
		 )
	      {
		// this one is good, but may leave mayhem elsewhere
		arcImpl_->junk_ = false;
		debug(true
		      ,"Serial ", lingaImpl_.intf_->track()
		      ," Arc of id ",arcImpl_->id_
		      ," unjunked"
		      );
	      }
	  }
      }
  }

  void
  ArcXMLparser::on_start_element(
				 const Glib::ustring & name
				 ,const xmlpp::SaxParser::AttributeList& attributes
				 )
  {
    ArcImpl & arcImpl = *arcImpl_;

    if ( name == "Metas" )
      {
	// a subelement of the latest arc
	new MetasXMLparser(top()		// this
			   ,arcImpl.intf_->metas()
			   );
      }
    else if ( name == "Label" )
      {
	new LabelXMLparser(*this
			   ,arcImpl
			   ,attributes
			   );
      }
    else
      throw XMLparser::ContextCorruption("LingA"
					 ,name
					 );
  }

  
  void
  ArcXMLparser::on_end_element(
			       const Glib::ustring& 	name
			       )
  {
    // now enter this arc into arcsByEpoch_, arcsByLabel_
    if ( arcImpl_->epochDefined() )
      lingaImpl_.arcsByEpoch_.insert(std::make_pair(Epoch(arcImpl_->left_->time_
							  ,arcImpl_->right_->time_
							  )
						    , arcImpl_
						    )
				     );
    lingaImpl_.arcsByLabel_.insert(std::make_pair(arcImpl_->label_
						  , arcImpl_
						  )
				   );
    
  }

  LabelXMLparser::LabelXMLparser(
				 ArcXMLparser &		arcXMLparser
				 , ArcImpl &		arcImpl
				 , const xmlpp::SaxParser::AttributeList& attributes
		   )
    : XMLparser("Label"
		,&arcXMLparser
		)
    , arcImpl_(arcImpl)
  {
    for ( xmlpp::SaxParser::AttributeList::const_iterator i = attributes.begin()
	    ;
	  i != attributes.end()
	    ;
	  ++i
	  )
      {
	if ( i->name == "randomVar" )
	  {
	    arcImpl_.label_.randomVar = i->value;
	  }
	else if ( i->name == "likelihood" )
	  {
	    arcImpl_.label_.likelihood = i->value;
	  }
      }
    if ( arcImpl_.label_.randomVar.empty() )
      {
	// make it up as unknown
	arcImpl_.label_.randomVar = "Unknown";
      }
    // empty label value so we can append to it
    arcImpl_.label_.value = "";
  }
  
  void
  LabelXMLparser::on_characters(
				const Glib::ustring& characters
				)
  {
    // 3-byte diacriticised IPAs are parsed as separate character!
    // This is a workaround
    arcImpl_.label_.value += characters;

    debug(true,
	  "LabelXMLparser: parsed label value = ", arcImpl_.label_.value.raw()
	  );
  }

}
