/*===========================================================================*/
/*
 * This file is part of
 * libalinga - a C++ library for Alligned Linguistic Annotation
 *
 * Copyright (C) 2007, 2008 Elaine Tsiang YueLien
 *
 * libalinga is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, see http://www.gnu.org/licenses, or write to the
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301, USA
 *
 *===========================================================================*/
/** @file LingAImpl.H                                                        *
 *                                                                           *
 *  ALingA::LingAImpl - LingA implementation interface                       *
 *                                                                           */
/*===========================================================================*/
#ifndef	LingAImplementationInterface
#define	LingAImplementationInterface

#ifdef __GNUG__
#pragma interface
#endif

#include	<Ogg/Logical.H>

namespace ALingA
{
  class LingAImpl;
  class NodeXMLparser;
  class ArcXMLparser;
  class LabelXMLparser;
}

#include	<ALingA/Signal.H>
#include	<ALingA/LingA.H>
#include	<NodesArcsImpl.H>
#include	<config.h>

#include	<map>

namespace	ALingA
{
  typedef std::map<Id, NodeImpl *>				NodesById;
  typedef NodesById::iterator					iNodesById;

  typedef std::map<Id, ArcImpl *>				ArcsById;
  typedef ArcsById::iterator					iArcsById;

  typedef std::multimap<Position, NodeImpl *>			NodesByTime;
  typedef NodesByTime::iterator					iNodesByTime;
  typedef NodesByTime::reverse_iterator				rNodesByTime;


  class PrecedingEpoch
  {
  public:
    bool
    operator()(const Epoch & e1, const Epoch & e2)
    {
      return(e1 < e2);
    }
  };
  
  typedef std::multimap<Epoch, ArcImpl *, PrecedingEpoch >	ArcsByEpoch;
  typedef ArcsByEpoch::iterator					iArcsByEpoch;
  typedef ArcsByEpoch::reverse_iterator				rArcsByEpoch;

  
  class PrecedingLabel
  {
  public:
    bool operator()(const Label & l1, const Label & l2) const
    {
      return(l1 < l2);
    }
  };

  typedef std::multimap<Label, ArcImpl *, PrecedingLabel >	ArcsByLabel;
  typedef ArcsByLabel::iterator					iArcsByLabel;

  class LingADriverParser;
  class LingAXMLparser;
  class NodeXMLparser;
  class ArcXMLparser;

  class LingAImpl /*, private ReadWriteLock
		    to be added later
		  */
  {
    friend class LingA;
    friend class Node;
    friend class NodeImpl;
    friend class Arc;
    friend class ArcImpl;

    friend class LingADriverParser;
    friend class LingAXMLparser;
    friend class NodeXMLparser;
    friend class ArcXMLparser;

  private:
    static
    const unsigned short	dtdMajor_ = ALINGA_DTD_MAJOR;
    static
    const unsigned short	dtdMinor_ = ALINGA_DTD_MINOR;


    LingA *			intf_;

    ALingA &			alinga_;

    long			signalTrack_;
    Signal const *		signal_;

    Metas 			metas_;

    Ogg::Logical::Writer *	writer_;
    XMLwriter *			xmlWriter_;
    XMLwriter *			lingaWriter_;
    NodeImpl *			lastNodeWritten_;

    LingADriverParser *		xmlParser_;
    Ogg::Logical::Reader *	reader_;

    // readwrite locks for each?
    NodesById			nodesById_;
    NodesByTime			nodesByTime_;
    ArcsById			arcsById_;
    ArcsByEpoch			arcsByEpoch_;
    ArcsByLabel			arcsByLabel_;

    bool			doneParsing_;

  public:
    LingAImpl(
	      ALingA &
	      ,long		signalTrack
	      );

    LingAImpl(
	      ALingA &
	      );

    virtual
    ~LingAImpl()
    {}

    void
    write();

    ArcImpl &
    arcImpl(iArcs &) const;

    void
    write(
	  XMLwriter &	nodeWriter
	  ,ArcImpl &
	  );

    Position
    basePosition(NodeImpl &);

    void
    write(
	  NodeImpl &
	  );

    void
    write(
	  Position	time
	  );

    void
    end();

    Id
    nextNodeId();

    Id
    nextArcId();
    
    void
    node(NodeImpl &);

    void
    erase(NodeImpl &);

    void
    arc(ArcImpl &);

    void
    erase(ArcImpl &);
    
    void
    readHeader();

    void
    readNode();

    void
    erase();

    bool
    operator==(LingAImpl &);

    Arcs
    arcsCrossing(Epoch & epoch
		 , bool includeContaining
		 );

    Nodes
    nodesIn(Epoch & epoch);

    Arcs
    sequenceIn(Epoch &
	       , Arc *
	       );

    Arcs
    sequenceContainedIn(Epoch &
			);

    Arcs
    includeArcInSequence(Arc &
			 , Arcs &
			 , Arc * hint
			 );
  };

  class LingADriverParser : public XMLparser
  {
    friend class	LingAImpl;
    friend class	XMLparser;

  private:
    LingAImpl &		lingaImpl_;

    LingADriverParser(
		 LingAImpl &	lingaImpl
		 )
      : XMLparser("Driver")
      , lingaImpl_(lingaImpl)
    {}

    ~LingADriverParser()
    {}

    /* base behaviour
       void
       on_start_document();

       void
       on_end_document();
    */

    void
    on_start_element(
		     const Glib::ustring& 	name
		     , const xmlpp::SaxParser::AttributeList& attributes
		     );
    
    void
    on_end_element(
		   const Glib::ustring& 	name
		   )
    {
      throw XMLparser::UnexpectedElement(name);
    }
    
    void
    on_characters(
		  const Glib::ustring& characters
		  )
    {
      if ( characters != "\n" )
	throw XMLparser::UnexpectedCharacters(characters);

    }
    
    void on_cdata_block(
			const Glib::ustring& text
			)
    {
      throw XMLparser::UnexpectedCData(text);
    }
    
  };

  class LingAXMLparser : public XMLparser
  {
    friend class	LingAImpl;
    friend class	LingADriverParser;

  private:
    LingAImpl &		lingaImpl_;

  public:
    LingAXMLparser(
		   LingADriverParser &		driverParser
		   , LingAImpl &		lingaImpl
		   , const xmlpp::SaxParser::AttributeList& attributes
		   )
      ;

    ~LingAXMLparser()
    {}


    void
    on_start_document()
    {
      throw XMLparser::UnexpectedDocument();
    }
    

    void
    on_end_document()
    {
      throw XMLparser::UnexpectedDocument();
    }
    


    void
    on_start_element(
		     const Glib::ustring& 	name
		     , const xmlpp::SaxParser::AttributeList& attributes
		     );


    void
    on_end_element(
		   const Glib::ustring& 	name
		   );


    void
    on_characters(
		  const Glib::ustring& characters
		  )
    {
      if ( characters != "\n" )
	throw XMLparser::UnexpectedCharacters(characters);
    }
    
    void on_cdata_block(
			const Glib::ustring& text
			)
    {
      throw XMLparser::UnexpectedCData(text);
    }
  };

  class NodeXMLparser : public XMLparser
  {
  private:
    LingAImpl &		lingaImpl_;
    NodeImpl *		nodeImpl_;

  public:
    NodeXMLparser(
		  LingAXMLparser &	lingaXmlparser
		  , LingAImpl &		lingaImpl
		  , const xmlpp::SaxParser::AttributeList& attributes
		  )
      ;

    ~NodeXMLparser()
    {}


    void
    on_start_document()
    {
      throw XMLparser::UnexpectedDocument();
    }
    

    void
    on_end_document()
    {
      throw XMLparser::UnexpectedDocument();
    }
    


    void
    on_start_element(
		     const Glib::ustring& 	name
		     , const xmlpp::SaxParser::AttributeList& attributes
		     );
    void
    on_end_element(
		   const Glib::ustring& 	name
		   );

    
    void
    on_characters(
		  const Glib::ustring& characters
		  )
    {
      if ( characters != "\n" )
	throw XMLparser::UnexpectedCharacters(characters);
    }
    
    void on_cdata_block(
			const Glib::ustring& text
			)
    {
      throw XMLparser::UnexpectedCData(text);
    }
  };

  class ArcXMLparser : public XMLparser
  {
  private:
    LingAImpl &		lingaImpl_;
    NodeImpl &		nodeImpl_;
    ArcImpl *		arcImpl_;

  public:
    ArcXMLparser(
		 NodeXMLparser &	nodeXmlParser
		 , LingAImpl &		lingaImpl
		 , NodeImpl &		nodeImpl
		 , const xmlpp::SaxParser::AttributeList& attributes
		 )
      ;

    ~ArcXMLparser()
    {}


    void
    on_start_document()
    {
      throw XMLparser::UnexpectedDocument();
    }
    

    void
    on_end_document()
    {
      throw XMLparser::UnexpectedDocument();
    }
    


    void
    on_start_element(
		     const Glib::ustring& 	name
		     , const xmlpp::SaxParser::AttributeList& attributes
		     );
    void
    on_end_element(
		   const Glib::ustring& 	name
		   );

    
    void
    on_characters(
		  const Glib::ustring& characters
		  )
    {
      if ( characters != "\n" )
	throw XMLparser::UnexpectedCharacters(characters);
    }
    
    void on_cdata_block(
			const Glib::ustring& text
			)
    {
      throw XMLparser::UnexpectedCData(text);
    }
  };

  class LabelXMLparser : public XMLparser
  {
  private:
    ArcImpl &		arcImpl_;

  public:
    LabelXMLparser(
		   ArcXMLparser &	arcXmlParser
		   , ArcImpl &		arcImpl
		   , const xmlpp::SaxParser::AttributeList& attributes
		   )
      ;

    ~LabelXMLparser()
    {}


    void
    on_start_document()
    {
      throw XMLparser::UnexpectedDocument();
    }
    

    void
    on_end_document()
    {
      throw XMLparser::UnexpectedDocument();
    }
    


    void
    on_start_element(
		     const Glib::ustring& 	name
		     , const xmlpp::SaxParser::AttributeList& attributes
		     )
    {
      throw XMLparser::UnexpectedElement(name);
    }
    
    /* base behavior
    void
    on_end_element(
		   const Glib::ustring& 	name
		   )
    {
    }
    */
    
    void
    on_characters(
		  const Glib::ustring& characters
		  );
    
    void on_cdata_block(
			const Glib::ustring& text
			)
    {
      throw XMLparser::UnexpectedCData(text);
    }
  };

}

#endif
