/*===========================================================================*/
/*
 * This file is part of
 * libalinga - a C++ library for Alligned Linguistic Annotation
 *
 * Copyright (C) 2007, 2008 Elaine Tsiang YueLien
 *
 * libalinga is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, see http://www.gnu.org/licenses, or write to the
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301, USA
 *
 *===========================================================================*/
/** @file ManifoldImpl.C                                                     *
 *                                                                           *
 *  ALingA::ManifoldImpl - Manifold implementation                           *
 *                                                                           */
/*===========================================================================*/
#ifdef __GNUG__
#pragma implementation "Manifold.H"
#pragma implementation "ManifoldImpl.H"
#endif

#include	<ManifoldImpl.H>

extern "C" 
{
# include	<stdlib.h>
}

namespace	ALingA
{
  Manifold::Manifold(
		     const size_t		perTime
		     ,const size_t		perSpace
		     ,const Glib::ustring	timeUnit
		     ,const Glib::ustring	spaceUnit
		     ,const Position		timeBias
		     ,const Position		spaceBias
		     ,const size_t		gransPerFrame
		     ,const Glib::ustring	timeOrigin
		     )
  {
    ManifoldImpl *impl = new ManifoldImpl(
					  perTime   
					  ,perSpace  
					  ,timeUnit  
					  ,spaceUnit 
					  ,timeBias  
					  ,spaceBias 
					  ,gransPerFrame
					  ,timeOrigin
					  );

    impl->intf_ = this;
    impl_ = impl;
  }

  ManifoldImpl::ManifoldImpl(
			     const size_t		perTime
			     ,const size_t		perSpace
			     ,const Glib::ustring	timeUnit
			     ,const Glib::ustring	spaceUnit
			     ,const Position		timeBias
			     ,const Position		spaceBias
			     ,const size_t		gransPerFrame
			     ,const Glib::ustring	timeOrigin
			     )
    : perTime_(perTime)
    , perSpace_(perSpace)
    , timeUnit_(timeUnit)
    , spaceUnit_(spaceUnit)
    , timeBias_(timeBias)
    , spaceBias_(spaceBias)
    , gransPerFrame_(gransPerFrame)
    , timeOrigin_(timeOrigin)
  {
  }

  Manifold::Manifold(
		     const Manifold &	manifold
		     )
  {
    const ManifoldImpl *impl0 = static_cast<const ManifoldImpl *>(manifold.impl());
    ManifoldImpl *impl  = new ManifoldImpl(
					   impl0->perTime_
					   , impl0->perSpace_
					   , impl0->timeUnit_  
					   , impl0->spaceUnit_ 
					   , impl0->timeBias_  
					   , impl0->spaceBias_ 
					   , impl0->gransPerFrame_
					   , impl0->timeOrigin_
					  );

    impl->intf_ = this;
    impl_ = impl;

  }

  Manifold::Manifold(
		     const Manifold & manifold
		     ,const Signal &	signal
		     )
  {
    ManifoldImpl *impl0 = static_cast<ManifoldImpl *>(manifold.impl());
    ManifoldImpl *impl  = new ManifoldImpl(
					  impl0->perTime_
					  , impl0->perSpace_
					  , impl0->timeUnit_  
					  , impl0->spaceUnit_ 
					  , signal.bias().time
					  , signal.bias().space 
					  , impl0->gransPerFrame_
					  , impl0->timeOrigin_
					  );

    impl->intf_ = this;

    impl->perTime_ = impl0->perTime_ / signal.decimation().time;
    impl->perSpace_ = impl0->perSpace_ / signal.decimation().space;
  }

  Manifold::Manifold(
		     const Attribute &	attribute
		     )
  {
    if ( attribute.name == "perTime" )
      {
	const long perTime = *(reinterpret_cast<long *>(attribute.typedValue()));

	ManifoldImpl *impl = new ManifoldImpl(
					      perTime
					      ,1	//perSpace
					      ,"Sec"	//timeUnit
					      ,""	//spaceUnit
					      ,0	//timeBias
					      ,0	//spaceBias
					      ,1	//gransPerFrame
					      ,""	//timeOrigin
					      );
	impl->intf_ = this;
	impl_ = impl;
      }
    else
      throw Manifold::LoneAttributeNotPerTime(attribute.name);
  }

  Manifold::~Manifold()
  {
    if ( impl_ )
      {
	ManifoldImpl * impl = static_cast<ManifoldImpl *>(impl_);
	delete impl;
	impl_ = 0;
      }
  }

  void
  Manifold::operator()(
		       const xmlpp::SaxParser::Attribute & attribute
		       )
  {
    ManifoldImpl * impl = static_cast<ManifoldImpl *>(impl_);

    if ( attribute.name == "perTime" )
      {
	long long ll = atoll(attribute.value.c_str());
	if ( ll <= 0 )
	  ll = 0;
	impl->perTime_ = ll;
      }
    else if ( attribute.name == "perSpace" )
      {
	long long ll = atoll(attribute.value.c_str());
	if ( ll <= 0 )
	  ll = 0;
	impl->perSpace_ = ll;
      }
    else if ( attribute.name == "timeUnit" )
      {
	impl->timeUnit_ = attribute.value;
      }
    else if ( attribute.name == "spaceUnit" )
      {
	impl->spaceUnit_ = attribute.value;
      }
    else if ( attribute.name == "timeBias" )
      {
	impl->timeBias_ = atoll(attribute.value.c_str());
      }
    else if ( attribute.name == "spaceBias" )
      {
	impl->spaceBias_ = atoll(attribute.value.c_str());
      }
    else if ( attribute.name == "gransPerFrame" )
      {
	long long ll = atoll(attribute.value.c_str());
	if ( ll < 1 )
	  ll = 0;
	impl->gransPerFrame_ = ll;
      }
    else if ( attribute.name == "timeOrigin" )
      {
	impl->timeOrigin_ = attribute.value;
      }
  }

  Manifold &
  Manifold::operator=(const Manifold & m)
  {
    ManifoldImpl * m0 = static_cast<ManifoldImpl *>(impl_);
    const ManifoldImpl * m1 = static_cast<const ManifoldImpl *>(m.impl_);

    m0->perTime_ = m1->perTime_;
    m0->perSpace_ = m1->perSpace_;
    m0->timeUnit_ = m1->timeUnit_;
    m0->spaceUnit_ = m1->spaceUnit_;
    m0->timeBias_ = m1->timeBias_;
    m0->spaceBias_ = m1->spaceBias_;
    m0->gransPerFrame_ = m1->gransPerFrame_;
    m0->timeOrigin_ = m1->timeOrigin_;

    return(*this);
  }

  Decimation
  Manifold::operator / (
			const Manifold & manifold
			) const
  {
    const ManifoldImpl * impl = static_cast<ManifoldImpl *>(impl_);
    const ManifoldImpl * impl1 = static_cast<ManifoldImpl *>(manifold.impl());

    if ( impl->timeUnit_ != impl1->timeUnit_ )
      throw Manifold::OrderIndeterminate(impl->timeUnit_
					 ,impl1->timeUnit_
					 );

    if ( impl->spaceUnit_ != impl1->spaceUnit_ )
      throw Manifold::OrderIndeterminate(impl->spaceUnit_
					 ,impl1->spaceUnit_
					 );

    if ( impl->timeOrigin_ != impl1->timeOrigin_ )
      throw Manifold::OrderIndeterminate(impl->timeOrigin_
					 ,impl1->timeOrigin_
					 );

    if ( ( impl->perTime_ % impl1->perTime_ )
	 != 0
	 )
      throw Manifold::NotBase(LongAttribute(
					    Glib::ustring("perTime")
					    ,impl->perTime_
					    )
			      ,LongAttribute(
					     Glib::ustring("perTime")
					     ,impl1->perTime_
					     )
			      );

    if ( ( impl->perSpace_ % impl1->perSpace_ )
	 != 0
	 )
      {
	  throw Manifold::NotBase(LongAttribute(
						Glib::ustring("perSpace")
						,impl->perSpace_
						)
				  ,LongAttribute(
						 Glib::ustring("perSpace")
						 ,impl1->perSpace_
						 )
				  );
      }

    Decimation	decimation(impl->perTime_ / impl1->perTime_
			   ,impl->perSpace_ / impl1->perSpace_
			   );

    if (
	impl->timeBias_ > impl1->timeBias_ 
	)
      throw Manifold::NotBase(LongLongAttribute(
						Glib::ustring("timeBias")
						,impl->timeBias_
						)
			      ,LongLongAttribute(
						 Glib::ustring("timeBias")
						 ,impl1->timeBias_
						 )
			      );

    if (
	impl->spaceBias_ > impl1->spaceBias_ 
	)
      throw Manifold::NotBase(LongLongAttribute(
						Glib::ustring("spaceBias")
						,impl->spaceBias_
						)
			      ,LongLongAttribute(
						 Glib::ustring("spaceBias")
						 ,impl1->spaceBias_
						 )
			      );

    Position topBaseGran = impl->spaceBias_ + impl->gransPerFrame_ - 1;
    Position topGran = impl1->spaceBias_ + (impl1->gransPerFrame_ - 1)* decimation.space;
    if ( topBaseGran < topGran )
      throw Manifold::NotBase(SizeAttribute(
					    Glib::ustring("top gran")
					    ,topBaseGran
					    )
			      ,SizeAttribute(
					     Glib::ustring("top gran")
					     ,topGran
					    )
			      );


    return(decimation);
  }

  bool
  Manifold::swap(Manifold & manifold)
  {
    bool swapped(false);

    if ( (perSpace() == 1)
	 && ((perSpace() % manifold.perSpace()) == 1)
	 )
      {
	// space swap only
	Manifold newMe(
		       perTime()
		       ,manifold.perSpace()
		       ,timeUnit()
		       ,spaceUnit()
		       ,timeBias()
		       ,spaceBias()
		       ,gransPerFrame()
		       ,timeOrigin()
		       );
	*this = newMe;
	swapped = true;
      }

    Decimation	d(1,1);
    Manifold	me(*this);

    try
      {
	d =  me / manifold;
      }
    catch ( Manifold::NotBase nb1 )
      {
	try
	  {
	    d = manifold / me;
	  }
	catch ( Manifold::NotBase nb2 )
	  {
	    throw Manifold::IncompatibleManifolds(nb1.what()
						  ,nb2.what()
						  );
	  }
	// full swap
	*this = manifold;
	swapped = true;
      }

    return(swapped);
  }

  Attributes
  Manifold::attributes() const
  {
    const ManifoldImpl * impl = static_cast<ManifoldImpl *>(impl_);

    Attributes	l;

    l.push_back(
		SizeAttribute(
			      Glib::ustring("perTime")
			      ,impl->perTime_
			      )
		);
    l.push_back(
		SizeAttribute(
			      Glib::ustring("perSpace")
			      ,impl->perSpace_
			      )
		);
    l.push_back(
		xmlpp::SaxParser::Attribute(
					    Glib::ustring("timeUnit")
					    ,impl->timeUnit_
					    )
		);
    l.push_back(
		xmlpp::SaxParser::Attribute(
					    Glib::ustring("spaceUnit")
					    ,impl->spaceUnit_
					    )
		);
    l.push_back(
		SizeAttribute(
			      Glib::ustring("timeBias")
			      ,impl->timeBias_
			      )
		);
    l.push_back(
		SizeAttribute(
			      Glib::ustring("spaceBias")
			      ,impl->spaceBias_
			      )
		);
    l.push_back(
		SizeAttribute(
			      Glib::ustring("gransPerFrame")
			      ,impl->gransPerFrame_
			      )
		);
    l.push_back(
		xmlpp::SaxParser::Attribute(
					    Glib::ustring("timeOrigin")
					    ,impl->timeOrigin_
					    )
		);
    return(l);
  }

  size_t
  Manifold::perTime() const
  {
    const ManifoldImpl * impl = static_cast<ManifoldImpl *>(impl_);

    return(impl->perTime_);
  }

  size_t
  Manifold::perSpace() const
  {
    const ManifoldImpl * impl = static_cast<ManifoldImpl *>(impl_);

    return(impl->perSpace_);
  }

  Glib::ustring
  Manifold::timeUnit() const
  {
    const ManifoldImpl * impl = static_cast<ManifoldImpl *>(impl_);

    return(impl->timeUnit_);
  }

  Glib::ustring
  Manifold::spaceUnit() const
  {
    const ManifoldImpl * impl = static_cast<ManifoldImpl *>(impl_);

    return(impl->spaceUnit_);
  }

  Position
  Manifold::timeBias() const
  {
    const ManifoldImpl * impl = static_cast<ManifoldImpl *>(impl_);

    return(impl->timeBias_);
  }

  Position
  Manifold::spaceBias() const
  {
    const ManifoldImpl * impl = static_cast<ManifoldImpl *>(impl_);

    return(impl->spaceBias_);
  }

  size_t
  Manifold::gransPerFrame() const
  {
    const ManifoldImpl * impl = static_cast<ManifoldImpl *>(impl_);

    return(impl->gransPerFrame_);
  }

  Glib::ustring
  Manifold::timeOrigin() const
  {
    const ManifoldImpl * impl = static_cast<ManifoldImpl *>(impl_);

    return(impl->timeOrigin_);
  }

  //--Exceptions---------------------------------------------------------------------------

  Manifold::OrderIndeterminate::OrderIndeterminate(
						   const Glib::ustring	first
						   ,const Glib::ustring	second
						   ) throw()
    : Exception("Order of manifolds not determinable because ")
  {
    *this << first << " is not equal to " << second << std::endl;
  }

  Manifold::NotBase::NotBase(
			     const ::ALingA::Attribute	first
			     ,const ::ALingA::Attribute	second
			     ) throw()
    : Exception("")
  {
    *this
      << first.name << " of " << first.value << " is not a base for "
      << second.name << " of " << second.value << std::endl;
  }

  Manifold::IncompatibleManifolds::IncompatibleManifolds(
							 const char *	what1
							 ,const char *	what2
							 ) throw()
    : Exception("Manifolds are incompatible because ")
  {
    *this
      << std::endl
      << what1 << std::endl
      << " and " << std::endl
      << what2 << std::endl;
  }

  Manifold::LoneAttributeNotPerTime::LoneAttributeNotPerTime(
							     const Glib::ustring attrnm
							     ) throw()
    : Exception("Lone attribute of ")
  {
    *this << attrnm << " is not perTime" << std::endl;
  }

  Manifold::InvalidAttribute::InvalidAttribute(
					       const Glib::ustring attrnm
					       ) throw()
    : Exception("Attribute of ")
  {
    *this << attrnm << " is not valid" << std::endl;
  }

}
