/*===========================================================================*/
/*
 * This file is part of
 * libalinga - a C++ library for Alligned Linguistic Annotation
 *
 * Copyright (C) 2007, 2008 Elaine Tsiang YueLien
 *
 * libalinga is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, see http://www.gnu.org/licenses, or write to the
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301, USA
 *
 *===========================================================================*/
/** @file NodesArcsImpl.C                                                    *
 *                                                                           *
 *  ALingA::Nodes ALingA::Arcs implementation                                *
 *                                                                           */
/*===========================================================================*/
#ifdef __GNUG__
#pragma implementation "NodesArcs.H"
#pragma implementation "NodesArcsImpl.H"
#endif

#include	<NodesArcsImpl.H>
#include	<Ogg/Debug.H>

namespace ALingA
{
  using namespace Ogg;

  /*-------Node related------------------------------------------------------------*/
  void *
  Node::impl()
  {
    return(impl_);
  }

  Node::Node(
	     LingA &		linga
	     ,Coordinate	co
	     )
  {
    LingAImpl & laImpl = *static_cast<LingAImpl *>(linga.impl_);
    
    NodeImpl * imp = new NodeImpl(
				  laImpl
				  ,laImpl.nextNodeId()
				  ,(long long)(co.time)
				  ,(long long)(co.space)
				  );
    imp->intf_ = this;
    impl_ = imp;
  }

  Node::Node(
	     LingA &	linga
	     ,Position	time
	     )
  {
    LingAImpl & laImpl = *static_cast<LingAImpl *>(linga.impl_);
    
    NodeImpl * impl = new NodeImpl(
				   laImpl
				   ,laImpl.nextNodeId()
				   ,(long long)(time)
				   ,(long long)(-1)
				   );
    impl->intf_ = this;
    impl_ = impl;
  }

  Node::Node(
	     LingA &	linga
	     )
  {
    LingAImpl & laImpl = *static_cast<LingAImpl *>(linga.impl_);
    
    NodeImpl * impl = new NodeImpl(
				   laImpl
				   ,laImpl.nextNodeId()
				   ,-1
				   ,-1
				   );
    impl->intf_ = this;
    impl_ = impl;
  }

  Node::Node(
	     Id	id
	     ,LingA &	linga
	     )
  {
    LingAImpl & laImpl = *static_cast<LingAImpl *>(linga.impl_);
    
    NodeImpl * impl = new NodeImpl(
				   laImpl
				   ,id
				   ,-1
				   ,-1
				   );
    impl->intf_ = this;
    impl_ = impl;
  }

  NodeImpl::NodeImpl(
		     LingAImpl &	linga
		     ,Id		id
		     ,long long		time
		     ,long long		space
		     )
    : intf_(0)
    , id_(id)
    , lefts_()
    , rights_()
    , time_(time)
    , space_(space)
    , linga_(linga)
    , written_(false)
    , junk_(false)
  {
    linga.node(*this);
    debug(true
	  ,"Serial ",linga_.intf_->track()
	  ," created Node of id ",id_
	  );
  }

  Node::~Node()
  {
    delete static_cast<NodeImpl *>(impl_);
    impl_ = 0;
  }


  NodeImpl::~NodeImpl()
  {
    // Must delete all attached Arcs!
    for ( iArcs i = lefts_.begin()
	    ;
	  i != lefts_.end()
	    ;
	  ++i
	  )
      {
	delete *i;
      }

    for ( iArcs i = rights_.begin()
	    ;
	  i != rights_.end()
	    ;
	  ++i
	  )
      {
	delete *i;
      }

    linga_.erase(*this);

  }

  Id
  Node::id() const
  {
    return( static_cast<NodeImpl *>(impl_)->id_ );
  }

  LingA &
  Node::linga()
  {
    return( *(static_cast<NodeImpl *>(impl_)->linga_).intf_ );
  }

  bool
  Node::time(Position time)
  {
    if ( !lefts().empty() )
      {
	// rbegin() is the shortest epoch
	iArcs ia = --lefts().end();
	try
	  {
	    Epoch((*ia)->left().time(), time);
	  }
	catch( Epoch::Invalid )
	  {
	    // if Epoch::Invalid, then don't do it
	    return(false);
	  }
	catch( Node::TimeUndefined )
	  {
	    // OK
	  }
      }
    if ( !rights().empty() )
      {
	// begin() is the shortest
	iArcs ia = rights().begin();
	try
	  {
	    Epoch(time, (*ia)->right().time());
	  }
	catch( Epoch::Invalid )
	  {
	    // if Epoch::Invalid, then don't do it
	    return(false);
	  }
	catch( Node::TimeUndefined )
	  {
	    // OK
	  }
      }

    NodeImpl * n = static_cast<NodeImpl *>(impl_);

    n->time(time);

    return true;
  }
  
  void
  NodeImpl::time(Position time)
  {
    // remove arcs from arcsBy*
    for ( iArcs i = lefts_.begin()
	    ;
	  i != lefts_.end()
	    ;
	  ++i
	  )
      {
	linga_.erase(*static_cast<ArcImpl *>((*i)->impl()));
      }

    for ( iArcs i = rights_.begin()
	    ;
	  i != rights_.end()
	    ;
	  ++i
	  )
      {
	linga_.erase(*static_cast<ArcImpl *>((*i)->impl()));
      }

    
    linga_.erase(*this);
    time_ = time;
    linga_.node(*this);

    // add arcs back into arcsBy*
    for ( iArcs i = lefts_.begin()
	    ;
	  i != lefts_.end()
	    ;
	  ++i
	  )
      {
	linga_.arc(*static_cast<ArcImpl *>((*i)->impl()));
      }

    for ( iArcs i = rights_.begin()
	    ;
	  i != rights_.end()
	    ;
	  ++i
	  )
      {
	linga_.arc(*static_cast<ArcImpl *>((*i)->impl()));
      }
  }

  bool
  Node::space(Position space)
  {
    // to be refined
    static_cast<NodeImpl *>(impl_)->space_ = space;
    return(true);
  }

  Position
  Node::time() const
  {
    if ( static_cast<NodeImpl *>(impl_)->time_ < 0 )
      throw TimeUndefined();
    return( static_cast<NodeImpl *>(impl_)->time_ );
  }

  Position
  Node::space() const
  {
    if ( static_cast<NodeImpl *>(impl_)->space_ < 0 )
      throw SpaceUndefined();
    return( static_cast<NodeImpl *>(impl_)->space_ );
  }

  Arcs
  Node::lefts() const
  {
    NodeImpl * n = static_cast<NodeImpl *>(impl_);
    
    debug(true
	  ,"Serial ", n->linga_.intf_->track()
	  ," node ",n->id_
	  ," has ",n->lefts_.size()
	  ," left arcs"
	  );
    return(n->lefts_); 
  }

  Arcs
  Node::rights() const
  {
    NodeImpl * n = static_cast<NodeImpl *>(impl_);
    
    debug(true
	  ,"Serial ", n->linga_.intf_->track()
	  ," node ",n->id_
	  ," has ",n->rights_.size()
	  ," right arcs"
	  );
    return(n->rights_); 
  }

  bool
  Node::junk()
  {
    return( static_cast<NodeImpl *>(impl_)->junk_ );
  }

  bool
  Node::operator==(const Node & node) const
  {
    return(
	   static_cast<NodeImpl *>(node.impl_)->id_
	   ==
	   static_cast<NodeImpl *>(impl_)->id_
	   );
  }

  bool
  NodeImpl::detached() {
    if ( lefts_.empty() && rights_.empty() )
      {
	debug(true,
	      "detached node ", id_
	      );
	return(true);
      }
    else
      return(false);
  }

  // time() and space() may throw Undefined
  bool
  operator< (const Node & n1, const Node & n2)
  {
    try
      {
	if ( n1.time() == n2.time() )
	  return( n1.space() < n2.space() );
	else
	  return( n1.time() < n2.time() );
      }
    catch ( Node::TimeUndefined )
      {
	NodeImpl & ni1 = *static_cast<NodeImpl *>(n1.impl_);
	NodeImpl & ni2 = *static_cast<NodeImpl *>(n2.impl_);
	return ( ni1.id_ < ni2.id_ );
      }
    catch ( Node::SpaceUndefined )
      {
	NodeImpl & ni1 = *static_cast<NodeImpl *>(n1.impl_);
	NodeImpl & ni2 = *static_cast<NodeImpl *>(n2.impl_);
	return ( ni1.id_ < ni2.id_ );
      }
  }

  Arcs
  path(Arc & arc1, Arc & arc2)
  {
    Arcs arcs;

    if ( arc1 == arc2 )
      {
	arcs.insert(&arc2);
	return arcs;
      }

    Arcs larcs(arc2.left().lefts());
    // Find a path from arc2 back to arc1
    for ( iArcs ia = larcs.begin()
	    ;
	  ia != larcs.end()
	    ;
	  ++ia
	  )
      {
	Arcs earlierPath = path(arc1, (**ia));
	if ( !earlierPath.empty() )
	  {
	    earlierPath.insert(&arc2);
	    return(earlierPath);
	  }
      }
    // return empty path
    return(arcs);
  }

  /*-------Arc related-------------------------------------------------------------*/
  void *
  Arc::impl()
  {
    return(impl_);
  }

  Arc::Arc(
	   Node &	left
	   ,Node &	right
	   ,Label	label
	   )
  {
    new(this) Arc(left
		  ,right
		  ,label
		  ,0
		  );
  }

  Arc::Arc(
	   Node &	left
	   ,Node &	right
	   ,Label	label
	   ,Id		id
	   )
  {
    try
      {
	Epoch(left.time(), right.time());
      }
    catch( Node::TimeUndefined )
      {
	debug(true
	      ," a node time undefined,"
	      ," continuing to create arc"
	      );
      }
    // if Epoch::Invalid, then let it sink

    LingAImpl & laImpl = *static_cast<LingAImpl *>(left.linga().impl_);
    NodeImpl & leftImpl = *(static_cast<NodeImpl*>(left.impl_));
    NodeImpl & rightImpl = *(static_cast<NodeImpl*>(right.impl_));
    
    if ( id <= 0 )
      {
	id = laImpl.nextArcId();

      }

    ArcImpl * imp = new ArcImpl(
				id
				,leftImpl
				,rightImpl
				,label
				,this
				);
    impl_ = imp;

    leftImpl.rights_.insert(this);
    debug(true
	  ,"Serial ",imp->linga_.intf_->track()
	  ," arc ", imp->id_
	  ,"'s left node ",left.id()," has "
	  ,left.rights().size()," right arcs"
	  );
    rightImpl.lefts_.insert(this);
    debug(true
	  ,"Serial ",imp->linga_.intf_->track()
	  ," arc ", imp->id_
	  ,"'s right node ",right.id()," has "
	  ,right.lefts().size()," left arcs"
	  );
  }

  ArcImpl::ArcImpl(
		   Id		id
		   ,NodeImpl &	left
		   ,NodeImpl &	right
		   ,Label	label
		   ,Arc *	arc
		   )
    : intf_(arc)
    , id_(id)
    , left_(&left)
    , right_(&right)
    , label_(label)
    , metas_(0)
    , linga_(left.linga_)
    , junk_(false)
  {
    if ( (0 <= right_->time_)
	 &&
	 (0 <= left_->time_)
	 )
      // test epoch
      Epoch(Position(left_->time_)
	    ,Position(right_->time_)
	    );

    linga_.arc(*this);
    debug(true
	  ,"Serial ",linga_.intf_->track()
	  ," created Arc of id ",id_
	  );

  }

  Arc::~Arc()
  {
    delete static_cast<ArcImpl *>(impl_);
    impl_ = 0;
  }

  ArcImpl::~ArcImpl()
  {
    detachLeft();
    detachRight();

    linga_.erase(*this);
  }

  void
  ArcImpl::detachLeft()
  {
    iArcs ia = left_->rights_.lower_bound(this->intf_);
    while ( ia != left_->rights_.end()
	    &&
	    *ia != this->intf_
	    )
      ++ia;
    if ( ia != left_->rights_.end() ) {     
      left_->rights_.erase(ia);
      debug(true
	    ,"Serial ",linga_.intf_->track()
	    ," arc of id ", id_
	    ," detached from left node of id ", left_->id_
	    );
    }
    
    // cannot deleteOnDetached because of specialized use during parsing
    // left_->deleteOnDetached();
  }

  void
  ArcImpl::detachRight()
  {
    iArcs ia = right_->lefts_.lower_bound(this->intf_);
    while ( ia != right_->lefts_.end()
	    &&
	    *ia != this->intf_
	    )
      ++ia;
    if ( ia != right_->lefts_.end() ) {
      right_->lefts_.erase(ia);
      debug(true
	    ,"Serial ",linga_.intf_->track()
	    ," arc of id ", id_
	    ," detached from right node of id ", right_->id_
	    );
    }

    // cannot deleteOnDetached because of specialized use during parsing
    // right_->deleteOnDetached();
  }

  // epoch() and box() may throw Undefined
  bool
  operator< (const Arc & arc1, const Arc & arc2)
  {
    try
      {
	if ( arc1.epoch() < arc2.epoch() )
	  return true;
	else if ( arc2.epoch() < arc1.epoch() )
	  return false;
	else
	  return ( arc1.box() < arc2.box() );
      }
    catch( Node::TimeUndefined )
      {
	ArcImpl & ai1 = *static_cast<ArcImpl *>(arc1.impl_);
	ArcImpl & ai2 = *static_cast<ArcImpl *>(arc2.impl_);
	return ( ai1.id_ < ai2.id_ );
      }
    catch( Node::SpaceUndefined )
      {
	ArcImpl & ai1 = *static_cast<ArcImpl *>(arc1.impl_);
	ArcImpl & ai2 = *static_cast<ArcImpl *>(arc2.impl_);
	return ( ai1.id_ < ai2.id_ );
      }
  }

  bool
  ArcImpl::epochDefined()
  {
    return((0 <= right_->time_)
	   &&
	   (0 <= left_->time_)
	   &&
	   (left_->time_ <= right_->time_)
	   );
  }

  Id
  Arc::id() const
  {
    return( static_cast<ArcImpl *>(impl_)->id_ );
  }

  Metas &
  Arc::metas()
  {
    ArcImpl & a = *static_cast<ArcImpl *>(impl_);

    if ( !a.metas_ )
      a.metas_ = new Metas;

    return(*a.metas_);
  }

  Epoch
  Arc::epoch() const
  {
    return(Epoch(left().time()
		 ,right().time()
		 )
	   );
  }

  Box
  Arc::box() const
  {
    return(Box(left().space()
	       ,right().space()
	       )
	   );
  }
  
  bool
  Arc::left(Node & node)
  {
    try
      {
	Epoch(node.time(), right().time());
      }
    catch( Epoch::Invalid )
      {
	// if Epoch::Invalid, then don't do it
	return(false);
      }
    catch( Node::TimeUndefined )
      {
	// OK
      }

    ArcImpl * a = static_cast<ArcImpl *>(impl_);
    NodeImpl * n = static_cast<NodeImpl *>(node.impl());
    NodeImpl * rn = a->right_; 

    if ( a->left_ != n ) {
      a->linga_.erase(*a);	//remove from arcsBy*
      a->detachRight();		//changing the left node affects the ordering of the left arcs of the right node!!!
      a->detachLeft();
      a->left_ = n;
      n->rights_.insert(this);
      a->right_ = rn;
      rn->lefts_.insert(this);
      debug(true
	    ,"Serial ",a->linga_.intf_->track()
	    ," arc of id ", a->id_
	    ," attached to left node of id ", n->id_
	    ," re-attached to right node of id ", rn->id_
	    );
      a->linga_.arc(*a);	//add back to arcsBy*
    }

    return(true);
  }

  Node &
  Arc::left() const
  {
    return(*(static_cast<ArcImpl *>(impl_)->left_->intf_) );
  }

  bool
  Arc::right(Node & node)
  {
    try
      {
	Epoch(left().time(), node.time());
      }
    catch( Epoch::Invalid )
      {
	// if Epoch::Invalid, then don't do it
	return(false);
      }
    catch( Node::TimeUndefined )
      {
	// OK
      }

    ArcImpl * a = static_cast<ArcImpl *>(impl_);
    NodeImpl * n = static_cast<NodeImpl *>(node.impl());
    NodeImpl * ln = a->left_; 

    if ( a->right_ != n )
      {
	a->linga_.erase(*a);	//remove from arcsBy*
	a->detachLeft();	//changing the right node affects the ordering of the right arcs of the left node!!!
	a->detachRight();
	a->right_ = n;
	n->lefts_.insert(this);
	a->left_ = ln;
	ln->rights_.insert(this);
	debug(true
	      ,"Serial ",a->linga_.intf_->track()
	      ," arc of id ", a->id_
	      ," attached to right node of id ", n->id_
	      ," re-attached to left node of id ", ln->id_
	      );
	a->linga_.arc(*a);	//add back to arcsBy*
      }

    return(true);
  }

  Node &
  Arc::right() const
  {
    return(*(static_cast<ArcImpl *>(impl_)->right_->intf_) );
  }

  void
  Arc::label(Label & lab)
  {
    ArcImpl * a = static_cast<ArcImpl *>(impl_);
    
    a->linga_.erase(*a);	//remove from arcsBy*
    static_cast<ArcImpl *>(impl_)->label_ = lab;
    a->linga_.arc(*a);		//add back to arcsBy*
  }

  Label
  Arc::label() const
  {
    return(static_cast<ArcImpl *>(impl_)->label_);
  }


  bool
  Arc::junk()
  {
    return( static_cast<ArcImpl *>(impl_)->junk_ );
  }

  bool
  Arc::operator==(const Arc & arc) const
  {
    return(
	   static_cast<ArcImpl *>(arc.impl_)->id_
	   ==
	   static_cast<ArcImpl *>(impl_)->id_
	   );
  }

}
