/*===========================================================================*/
/*
 * This file is part of
 * libalinga - a C++ library for Alligned Linguistic Annotation
 *
 * Copyright (C) 2007, 2008 Elaine Tsiang YueLien
 *
 * libalinga is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, see http://www.gnu.org/licenses, or write to the
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301, USA
 *
 *===========================================================================*/
/** @file SignalImpl.C                                                       *
 *                                                                           *
 *  ALingA::SignalImpl - Signal implementation                               *
 *                                                                           */
/*===========================================================================*/
#ifdef __GNUG__
#pragma implementation "Signal.H"
#pragma implementation "SignalImpl.H"
#endif

#include	<SignalImpl.H>
#include	<Ogg/Debug.H>

namespace	ALingA
{
  using Ogg::Debug;
  using Ogg::debug;

  long
  Signal::track() const
  {
    return(static_cast<const SignalImpl *>(this)->track_);
  }

  size_t
  Signal::baseGransPerFrame() const
  {
    return(static_cast<const SignalImpl *>(this)->baseGransPerFrame_);
  }

  Coordinate
  Signal::bias() const
  {
    return(static_cast<const SignalImpl *>(this)->bias_);
  }

  Decimation
  Signal::decimation() const
  {
    return(static_cast<const SignalImpl *>(this)->decimation_);
  }

  void
  Signal::href(Glib::ustring	hrf)
  {
    SignalImpl * sig = static_cast<SignalImpl *>(this);

    sig->href_ = hrf;
  }

  const Glib::ustring &
  Signal::href() const
  {
    return(static_cast<const SignalImpl *>(this)->href_);
  }

  void
  Signal::meta(MetaName		name
	       ,MetaValue	value
	       )
  {
    SignalImpl * sig = static_cast<SignalImpl *>(this);

    sig->metas_[name] = value;
  }

  const MetaValue &
  Signal::meta(const MetaName		name)
  {
    SignalImpl & sig = *static_cast<SignalImpl *>(this);
    
    const iMetas i = sig.metas_.find(name);
    if ( i != sig.metas_.end() )
      return(i->second);
    else
      throw SignalHasNoSuchMeta(name);
  }

  void
  SignalImpl::codecTypeFilled()
  {
    pthread_mutex_lock(&lockType_);
    {
      // wait for it to fill
      while( codecType_.empty() )
	pthread_cond_wait(&condType_, &lockType_);
    }
    pthread_mutex_unlock(&lockType_);
  }

  const Glib::ustring &
  Signal::codecType() const
  {
    return(static_cast<const SignalImpl *>(this)->codecType_);
  }

  void
  SignalImpl::codecTypeEmpty()
  {
    pthread_mutex_lock(&lockType_);
    {
      // wait for it to empty
      while( !codecType_.empty() )
	pthread_cond_wait(&condType_, &lockType_);
    }
    pthread_mutex_unlock(&lockType_);
  }

  void
  SignalImpl::codecType(
			Glib::ustring codecType
			)
  {
    pthread_mutex_lock(&lockType_);
    {
      if ( codecType_.empty() )
	codecType_ = codecType;
      else
	codecType_.clear();
      pthread_cond_signal(&condType_);
    }
    pthread_mutex_unlock(&lockType_);
  }

  void
  Signal::beginWrite(
		     Glib::ustring	codecType
		     )
  {
    SignalImpl * sig = static_cast<SignalImpl *>(this);

    if ( codecType.empty() )
      throw NullCodecType();

    if ( sig->codecType_.empty() )
      sig->codecType(
		     codecType
		     );
  }

  void
  Signal::endWrite()
  {
    SignalImpl * sig = static_cast<SignalImpl *>(this);

    if ( !sig->codecType_.empty() )
      // this call will clear it
      sig->codecType(
		     sig->codecType_
		     );

    debug(true,
	  "Track ", sig->track_
	  ," Signal should have declared it ended writing"
	  );

  }

  void
  Signal::endRead()
  {
    endWrite();
  }

  Coordinate
  Signal::baseCoordinate(Coordinate signalCoordinate) const
  {
    const SignalImpl * sig = static_cast<const SignalImpl *>(this);
    return sig->baseCoordinate(signalCoordinate);
  }

  Coordinate
  SignalImpl::baseCoordinate(Coordinate signalCoordinate) const
  {
    Position space  = bias_.space + signalCoordinate.space * decimation_.space;
    if ( space >= baseGransPerFrame_ )
      throw SpaceCoordinateOutOfBase(signalCoordinate.space
				     ,baseGransPerFrame_-1
				     );

    Position time  = bias_.time + signalCoordinate.time * decimation_.time;

    return(Coordinate(time,space));
  }

  Coordinate
  Signal::signalCoordinate(Coordinate baseCoordinate) const
  {
    const SignalImpl * sig = static_cast<const SignalImpl *>(this);
    return sig->signalCoordinate(baseCoordinate);
  }

  Coordinate
  SignalImpl::signalCoordinate(Coordinate baseCoordinate) const
  {
    Position time  = (baseCoordinate.time - bias_.time)  / decimation_.time;
    Position space = (baseCoordinate.space - bias_.space)/ decimation_.space;

    return(Coordinate(time,space));
  }
    
  Position
  Signal::basePosition(
		       Coordinate	signalCoordinate
		       ) const
  {
    const SignalImpl * sig = static_cast<const SignalImpl *>(this);
    return sig->basePosition(signalCoordinate);
  }

  Position
  SignalImpl::basePosition(
			   Coordinate	signalCoordinate
			   ) const
  {
    Coordinate coord = baseCoordinate(signalCoordinate);

    return( coord.time * baseGransPerFrame_
	    + coord.space
	    );
  }

  Position
  Signal::basePosition(Position		time
		       ,Position	space
		       ) const
  {
    return(basePosition(Coordinate(time,space)));
  }
}
