/*===========================================================================*/
/*
 * This file is part of
 * libalinga - a C++ library for Alligned Linguistic Annotation
 *
 * Copyright (C) 2007, 2008 Elaine Tsiang YueLien
 *
 * libalinga is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, see http://www.gnu.org/licenses, or write to the
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301, USA
 *
 *===========================================================================*/
/** @file XMLparserImpl.C                                                    *
 *                                                                           *
 *  ALingA::XMLparserImpl - XMLparser implementation                         *
 *                                                                           */
/*===========================================================================*/
#ifdef __GNUG__
#pragma implementation "XMLparser.H"
#pragma implementation "XMLparserImpl.H"
#endif

#include	<XMLparserImpl.H>
#include	<Ogg/Debug.H>

namespace	ALingA
{
  XMLparser::XMLparser(
		       Glib::ustring	elementName
		       ,XMLparser * 	parent
		       )
  {
    XMLparserImpl * parentImpl = 0;
    
    if ( parent )
      parentImpl = static_cast<XMLparserImpl *>(parent->impl_);

    XMLparserImpl * impl = new XMLparserImpl(
					     elementName
					     ,parentImpl
					     );
    impl->intf_ = this;
    impl_ = impl;
  }


  XMLparserImpl::XMLparserImpl(
			       Glib::ustring		elementName
			       ,XMLparserImpl * 	parent
			       )
    : SaxParser()
    ,elementName_(elementName)
  {
    if ( !parent )
      {
	context_ = new std::stack<XMLparserImpl *>;
	SaxParser::set_validate();
	SaxParser::set_substitute_entities();
      }
    else
      {
	context_ = parent->context_;
      }
    context_->push(this);
  }

  XMLparser::~XMLparser()
  {
    if ( impl_ )
      {
	XMLparserImpl * impl = static_cast<XMLparserImpl *>(impl_);
	delete impl;
	impl_ = 0;
      }
  }

  XMLparserImpl::~XMLparserImpl()
  {
    if ( context_ )
      {
	if ( !context_->empty() )
	  {
	    if ( context_->top() == this )
	      {
		context_->pop();
	      }
	    // otherwise do nothing
	    // on the way to crash
	  }
	if ( context_->empty() )
	  {
	    finish_chunk_parsing();
	    delete context_;
	  }
      }
    else
      throw XMLparser::NoContext(elementName_);
  }

  const Glib::ustring &
  XMLparser::elementName() const
  {
    XMLparserImpl * impl = static_cast<XMLparserImpl *>(impl_);

    return(impl->elementName_);
  }

  // Only the off-stack XMLparser ever calls parse
  // but the on-stack DERIVED XMLparsers' on_* get called
  // by the off-stack XMLparser's on_*
  void
  XMLparser::parse(
		   void *	data
		   ,size_t	size
		   )
  {
    XMLparserImpl * impl = static_cast<XMLparserImpl *>(impl_);

    // the data should be encoded as UTF8
    // first we make an ordinary string out of the data
    std::string s(
		  static_cast<char *>(data)
		  ,size
		  );
    // then we make a ustring from that
    Glib::ustring us(s);
    Ogg::debug(true
	       ,"XMLparser::parse parsing \n"
	       ,us.raw()
	       );
	  
    impl->parse_chunk(
		      us
		      );
    
    // parse_chunk ultimately calls impl->on_*
  }

  XMLparser &
  XMLparser::top()
  {
    XMLparserImpl * impl = static_cast<XMLparserImpl *>(impl_);

    XMLparserImpl * topimpl = impl->context_->top();
    
    return(*topimpl->intf_);
  }

  void
  XMLparser::on_start_document()
  {
    XMLparserImpl * impl = static_cast<XMLparserImpl *>(impl_);

    // the base behaviour
    impl->SaxParser::on_start_document();
  }

  void
  XMLparser::on_end_document()
  {
    XMLparserImpl * impl = static_cast<XMLparserImpl *>(impl_);

    // the base behaviour
    impl->SaxParser::on_end_document();
  }

  void
  XMLparser::on_start_element(
			      const Glib::ustring& name
			      , const xmlpp::SaxParser::AttributeList& attributes
			      )
  {
    XMLparserImpl * impl = static_cast<XMLparserImpl *>(impl_);

    // the base behaviour
    impl->SaxParser::on_start_element(
				      name
				      ,attributes
				      );
  }

  void
  XMLparser::on_end_element(const Glib::ustring& name)
  {
    XMLparserImpl * impl = static_cast<XMLparserImpl *>(impl_);

    // the base behaviour
    impl->SaxParser::on_end_element(name);
  }

  void
  XMLparser::on_characters(const Glib::ustring& characters)
  {
    XMLparserImpl * impl = static_cast<XMLparserImpl *>(impl_);

    // the base behaviour
    impl->SaxParser::on_characters(characters);
    
  }

  void
  XMLparser::on_cdata_block(const Glib::ustring& characters)
  {
    XMLparserImpl * impl = static_cast<XMLparserImpl *>(impl_);

    // the base behaviour
    impl->SaxParser::on_cdata_block(characters);
    
  }

  //--Exceptions---------------------------------------------------------------------------

  XMLparser::UnexpectedDocument::UnexpectedDocument() throw()
    : Exception("Unexpected Document ")
  {
  }

  XMLparser::UnexpectedElement::UnexpectedElement(
						  const Glib::ustring nm
						  ) throw()
    : Exception("Unexpected element ")
  {
    *this << nm << std::endl;
  }

  XMLparser::UnexpectedAttribute::UnexpectedAttribute(
						      const Glib::ustring nm
						      ) throw()
    : Exception("Unexpected attribute ")
  {
    *this << nm << std::endl;
  }

  XMLparser::UnexpectedCharacters::UnexpectedCharacters(
							const Glib::ustring nm
							) throw()
    : Exception("Unexpected characters ")
  {
    *this << nm << std::endl;
  }

  XMLparser::UnexpectedCData::UnexpectedCData(
					      const Glib::ustring nm
					      ) throw()
    : Exception("Unexpected CDATA ")
  {
    *this << nm << std::endl;
  }

  XMLparser::ContextCorruption::ContextCorruption(
						  const Glib::ustring nm
						  ,const Glib::ustring ctx
						  ) throw()
    : Exception("Context for ")
  {
    *this << nm << " corrupted to " << ctx << std::endl;
  }

  XMLparser::NoContext::NoContext(
				  const Glib::ustring nm
				  ) throw()
    : Exception("XMLparser bug! No context for ")
  {
    *this << nm << std::endl;
  }

  XMLparser::BadAttributeValue::BadAttributeValue(
						  const xmlpp::SaxParser::Attribute & attr
						  ) throw()
    : Exception("Bad Value ")
  {
    *this << attr.value << " for attribute " << attr.name << std::endl;
  }
}
