/*===========================================================================*/
/*
 * This file is part of
 * libalinga - a C++ library for Alligned Linguistic Annotation
 *
 * Copyright (C) 2007, 2008 Elaine Tsiang YueLien
 *
 * libalinga is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, see http://www.gnu.org/licenses, or write to the
 * Free Software Foundation, Inc.
 * 51 Franklin Street, Fifth Floor
 * Boston, MA  02110-1301, USA
 *
 *===========================================================================*/
/** @file XMLwriterImpl.C                                                    *
 *                                                                           *
 *  ALingA::XMLwriterImpl - XMLwriter implementation                         *
 *                                                                           */
/*===========================================================================*/
#ifdef __GNUG__
#pragma implementation "XMLwriter.H"
#pragma implementation "XMLwriterImpl.H"
#endif

#include	<XMLwriterImpl.H>
#include	<ALingA/Encode.H>

extern "C" 
{
# include	<string.h>
}


namespace	ALingA
{
  using		std::ostringstream;

  XMLwriter::XMLwriter(
		       Glib::ustring	rootName
		       ,unsigned short	dtdMajor
		       ,unsigned short	dtdMinor
		       )
  {
    XMLwriterImpl * impl = new XMLwriterImpl(
					     rootName
					     ,dtdMajor
					     ,dtdMinor
					     );
    impl->intf_ = this;
    impl_ = impl;
  }


  XMLwriterImpl::XMLwriterImpl(
			       Glib::ustring	rootName
			       ,unsigned short	dtdMajor
			       ,unsigned short	dtdMinor
			       )
    : dtdMajor_("dtdMajor", dtdMajor)
    , dtdMinor_("dtdMinor", dtdMinor)
    , elementName_(rootName)
    , utext_(*(new Glib::ustring()))
    , context_(new std::stack<XMLwriterImpl *>)
    , isInstruction_(false)
    , tagCompleted_(false)
  {
    if ( rootName.bytes() > 8 )
      throw XMLwriter::RootNameMoreThan8Bytes(rootName);

    // invariant
    utext_ += "<?xml version=\"1.0\" encoding=\"UTF8\" standalone=\"no\"?>\n";
    // VARIANT root name and SYSTEM
    utext_ += "<!DOCTYPE ";
    utext_ += rootName;
    utext_ += " SYSTEM \"";
    utext_ += rootName;
    utext_ += ".dtd\">\n";
  }

  XMLwriter::XMLwriter(
		       XMLwriter &	parent
		       ,Glib::ustring	elementName
		       ,bool		isInstruction
		       )
  {
    XMLwriterImpl * parentImpl = static_cast<XMLwriterImpl *>(parent.impl_);
    
    XMLwriterImpl * impl = new XMLwriterImpl(
					     *parentImpl
					     ,elementName
					     ,isInstruction
					     );
    impl->intf_ = this;
    impl_ = impl;
    
  }

  XMLwriterImpl::XMLwriterImpl(
			       XMLwriterImpl &	parent
			       ,Glib::ustring	elementName
			       ,bool		isInstruction
			       )
    : dtdMajor_(parent.dtdMajor_)
    , dtdMinor_(parent.dtdMinor_)
    , elementName_(elementName)
    , utext_(parent.utext_)
    , context_(parent.context_)
    , isInstruction_(isInstruction)
    , tagCompleted_(false)
  {
    
    utext_ += "<";
    if ( isInstruction_ )
      utext_ += "?";
    utext_ += elementName;
    utext_ += "\n";
    tagCompleted_ = false;

    context_->push(this);
  }

  XMLwriter::~XMLwriter()
  {
    if ( impl_ )
      {
	XMLwriterImpl * impl = static_cast<XMLwriterImpl *>(impl_);
	delete impl;
	impl_ = 0;
      }
  }

  XMLwriterImpl::~XMLwriterImpl()
  {
    if ( context_ )
      {
	if ( context_->empty() )
	  delete context_;
	else
	  if ( context_->top() == this )
	    {
	      if ( !isInstruction_ )
		{
		  if ( !tagCompleted_ )
		    throw XMLwriter::TagNotCompleted(elementName_);

		  utext_ += "</";
		  utext_ += elementName_;
		  utext_ += ">\n";
		}
	      context_->pop();
	    }
	  else
	    throw XMLwriter::ContextCorruption(elementName_);
      }
    else
      throw XMLwriter::NoContext(elementName_);
  }

  void
  XMLwriter::attribute(
		       Glib::ustring	name
		       ,Glib::ustring	value
		       )
  {
    XMLwriterImpl * impl = static_cast<XMLwriterImpl *>(impl_);

    impl->utext_ += "    ";
    impl->utext_ += name;
    impl->utext_ += "=\"";
    impl->utext_ += value;
    impl->utext_ += "\"\n";
  }

  void
  XMLwriter::endAttributes()
  {
    XMLwriterImpl * impl = static_cast<XMLwriterImpl *>(impl_);

    impl->utext_ += "    ";
    if ( impl->isInstruction_ )
      impl->utext_ += "?";
    impl->utext_ += ">";

    impl->tagCompleted_ = true;
  }

  void
  XMLwriter::characters(
			Glib::ustring &	characters
			)
  {
    XMLwriterImpl * impl = static_cast<XMLwriterImpl *>(impl_);

    impl->utext_ += characters;
  }

  void
  XMLwriter::manifold(
		      Manifold &	manifold
		      )
  {
    Attributes attributes(manifold.attributes());

    for ( Attributes::const_iterator i = attributes.begin()
	    ;
	  i != attributes.end()
	    ;
	  ++i
	  )
      {
	attribute(
		  i->name
		  ,i->value
		  );
      }
  }

  void
  XMLwriter::xmlns(
		   Glib::ustring	ns
		   )
  {
    XMLwriterImpl * impl = static_cast<XMLwriterImpl *>(impl_);

    attribute(
	      "dtdVersion"
	      ,impl->dtdMajor_.value + "." + impl->dtdMinor_.value
	      );
    attribute(
	      "xmlns"
	      ,ns
	      );
    attribute(
	      "xmlns:xlink"
	      ,"http://www.w3.org/1999/xlink"
	      );
  }

  void
  XMLwriter::metas(
		   Metas &	metas
		   )
  {
    for ( iMetas i = metas.begin()
	    ;
	  i != metas.end()
	    ;
	  ++i
	  )
      {
	XMLwriter	metaWriter(
				   *this
				   ,"Meta"
				   );
	metaWriter.attribute(
			     "name"
			     ,i->first	// name of Meta
			     );
	metaWriter.endAttributes();
	metaWriter.characters(i->second);
      }
  }

  size_t
  XMLwriter::textSize()
  {
    XMLwriterImpl * impl = static_cast<XMLwriterImpl *>(impl_);

    return( impl->utext_.bytes() );
  }

  void
  XMLwriter::write(
		   void *	data
		   )
  {
    XMLwriterImpl * impl = static_cast<XMLwriterImpl *>(impl_);

    memcpy(data
	   ,impl->utext_.data()
	   ,impl->utext_.bytes()
	   );
    impl->utext_ = "";
  }

  //--Exceptions---------------------------------------------------------------------------

  XMLwriter::RootNameMoreThan8Bytes::RootNameMoreThan8Bytes(
							    const Glib::ustring nm
							    ) throw()
    : Exception("Root name of ")
  {
    *this << nm << " is more than 8 bytes " << std::endl;
  }

  XMLwriter::TagNotCompleted::TagNotCompleted(
					      const Glib::ustring nm
					      ) throw()
    : Exception("Last tag of ")
  {
    *this << nm << " not completed" << std::endl;
  }

  XMLwriter::ContextCorruption::ContextCorruption(
						  const Glib::ustring nm
						  ) throw()
    : Exception("Context for ")
  {
    *this << nm << " corrupted" << std::endl;
  }

  XMLwriter::NoContext::NoContext(
				  const Glib::ustring nm
				  ) throw()
    : Exception("No context for ")
  {
    *this << nm << std::endl;
  }

}
